#!/usr/bin/env python3
""" Shared dustmaker CLI entrypoint """
import argparse
import sys

from dustmaker.cmd.thumbnail import Thumbnail
from dustmaker.cmd.transform import Transform
from dustmaker.cmd.variables import Variables

UTILITIES = (
    Thumbnail,
    Transform,
    Variables,
)


def main() -> int:
    """Shared CLI entrypoint for all dustmaker utilities"""
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="utility", help="which utility to use")

    util_map = {}
    for utility in UTILITIES:
        util = utility()  # type: ignore
        util_key = utility.__name__.lower()
        util.setup_parser(subparsers.add_parser(util_key))
        util_map[util_key] = util

    args = parser.parse_args()
    if args.utility is None:
        parser.print_help()
        return 0

    util = util_map[args.utility]
    del args.utility
    sys.exit(util.main(args))


if __name__ == "__main__":
    sys.exit(main())



"""
Module defining the tile respresentation in dustmaker.
"""
import copy
from dataclasses import dataclass
from enum import IntEnum
import math
from typing import Generator, Optional, Tuple

from .transform import TxMatrix


class TileSpriteSet(IntEnum):
    """Used to describe what set of tiles a tile's sprite comes from."""

    NONE_0 = 0
    MANSION = 1
    FOREST = 2
    CITY = 3
    LABORATORY = 4
    TUTORIAL = 5
    NEXUS = 6
    NONE_7 = 7


class TileSide(IntEnum):
    """Used to index the sides of a tile. This is the indexing done by the
    Dustforce engine itself."""

    TOP = 0
    BOTTOM = 1
    LEFT = 2
    RIGHT = 3


@dataclass
class TileEdgeData:
    """Data class for data stored on each tile edge. Many attributes are stored
    as pairs of data to correspond to the two corners of the tile edge. These
    corners are ordered clockwise around the tile (the tile should be on your
    right as you traverse from the first to second corner).
    """

    #: Should this edge produce collisions
    solid: bool = False
    #: Is this edge visible
    visible: bool = False
    #: Wether an edge cap should be drawn for each corner
    caps: Tuple[bool, bool] = (False, False)
    #: Edge join angle in degrees, should be in the range #: -90 < angle < 90.
    #: Ignored if the corresponding cap flag is set.
    angles: Tuple[int, int] = (0, 0)
    #: Sprite set of dust or spikes on this edge. Use `TileSpriteSet.NONE_0`
    #: to indicate no filth on this edge.
    filth_sprite_set: TileSpriteSet = TileSpriteSet.NONE_0
    #: If :attr:`filth_sprite_set` is not `TileSpriteSet.NONE_0` this flag
    #: controls if there is dust or spikes on the edge.
    filth_spike: bool = False
    #: Same as :attr:`caps` but for drawing filth (dust/spikes) caps.
    filth_caps: Tuple[bool, bool] = (False, False)
    #: Same as :attr:`angles` but for filth join angles.
    filth_angles: Tuple[int, int] = (0, 0)


class TileShape(IntEnum):
    """Tiles come in four main types; full, half, big, and small. Images of
    those tiles can be seen below. Alternatively refer to
    https://github.com/cmann1/PropUtils/blob/master/files/tiles_reference/TileShapes.jpg
    for an image of all tiles in one place.
    """

    #: .. image:: images/tiles/full.png
    FULL = 0
    #: .. image:: images/tiles/big_1.png
    BIG_1 = 1
    #: .. image:: images/tiles/small_1.png
    SMALL_1 = 2
    #: .. image:: images/tiles/big_2.png
    BIG_2 = 3
    #: .. image:: images/tiles/small_2.png
    SMALL_2 = 4
    #: .. image:: images/tiles/big_3.png
    BIG_3 = 5
    #: .. image:: images/tiles/small_3.png
    SMALL_3 = 6
    #: .. image:: images/tiles/big_4.png
    BIG_4 = 7
    #: .. image:: images/tiles/small_4.png
    SMALL_4 = 8
    #: .. image:: images/tiles/big_5.png
    BIG_5 = 9
    #: .. image:: images/tiles/small_5.png
    SMALL_5 = 10
    #: .. image:: images/tiles/big_6.png
    BIG_6 = 11
    #: .. image:: images/tiles/small_6.png
    SMALL_6 = 12
    #: .. image:: images/tiles/big_7.png
    BIG_7 = 13
    #: .. image:: images/tiles/small_7.png
    SMALL_7 = 14
    #: .. image:: images/tiles/big_8.png
    BIG_8 = 15
    #: .. image:: images/tiles/small_8.png
    SMALL_8 = 16
    #: .. image:: images/tiles/half_a.png
    HALF_A = 17
    #: .. image:: images/tiles/half_b.png
    HALF_B = 18
    #: .. image:: images/tiles/half_c.png
    HALF_C = 19
    #: .. image:: images/tiles/half_d.png
    HALF_D = 20


class Tile:
    """Represents a single tile in a Dustforce level. Positional information
    (x, y, layer) is stored within the containing :class:`Level` and not in
    the Tile object itself.

    Tiles support the equality and hashing interface.

    The constructor will by default create a square virtual tile with 4
    zeroed (non-solid nor visible) edges.

    Attributes:
        shape (TileShape): The shape of this particular tile.
        tile_flags (3-bit int): Raw bitmask of flags from the Dustforce engine.
            In practice this always seems to be 0x4 (which is set by default)
            corresponding to just the "solid" flag set. In most cases you should
            just ignore this flag. From the engine the definitions are:

            * Bit 1 - "solid slope flag" (probably ignored)
            * Bit 2 - "visible flag" (probably ignored)
            * Bit 3 - solid flag
        edge_data: List[TileEdgeData]: Edge data for each edge of the tile. This should always
            be a list of length 4 regardless of the tile :attr:`shape`.
        sprite_set (TileSpriteSet): The sprite set this tile comes from. (e.g.
            forest, mansion)
        sprite_tile (int): The index of the specific tile within this sprite set (e.g.
            grass, dirt). Check https://github.com/cmann1/PropUtils/tree/master/files/tiles_reference
            for a visual reference to get sprite index information.
        sprite_palette (int): The color variant of this tile.
    """

    def __init__(
        self,
        shape: TileShape = TileShape.FULL,
        *,
        tile_flags: int = 0x4,
        sprite_set: TileSpriteSet = TileSpriteSet.TUTORIAL,
        sprite_tile: int = 1,
        sprite_palette: int = 0,
        _tile_data: Optional[bytes] = None,
        _dust_data: Optional[bytes] = None,
    ) -> None:
        self.shape = shape
        self.tile_flags = tile_flags
        self.edge_data = [TileEdgeData() for _ in TileSide]
        self.sprite_set = sprite_set
        self.sprite_tile = sprite_tile
        self.sprite_palette = sprite_palette

        if _tile_data is not None:
            self._unpack_tile_data(_tile_data)
        if _dust_data is not None:
            self._unpack_dust_data(_dust_data)

    def __eq__(self, oth):
        if not isinstance(oth, Tile):
            return False
        return self._vals() == oth._vals()

    def __hash__(self):
        return hash(self._vals())

    def _vals(self):
        """Values used for __eq__ and __hash__"""
        return (
            self.shape,
            self.tile_flags,
            self.edge_data,
            self.sprite_set,
            self.sprite_tile,
            self.sprite_palette,
        )

    def get_sprite_tuple(self) -> Tuple[TileSpriteSet, int, int]:
        """Convenience method for getting a tuple that describes the sprite
        of a tile for easy sprite comparison and copying.

        Returns:
            A three-tuple containing the sprite set, tile, and palette of this tile.
        """
        return self.sprite_set, self.sprite_tile, self.sprite_palette

    def set_sprite_tuple(self, sprite_tuple: Tuple[TileSpriteSet, int, int]) -> None:
        """Convenience method for setting sprite information in the same format
        as :meth:`get_sprite_tuple`.

        Arguments:
            sprite_tuple (TileSpriteSet, int, int): Sprite set, tile, and palette
                information.
        """
        self.sprite_set, self.sprite_tile, self.sprite_palette = sprite_tuple

    @property
    def sprite_path(self) -> str:
        """Gives the path within the extracted sprite metadata to the tile sprite
        currently selected. You may retrieve the complete game sprites listing from
        https://www.dropbox.com/s/jm37ew9p74olgca/sprites.zip?dl=0
        """
        return "area/{}/tiles/tile{}_{}_0001.png".format(
            self.sprite_set.name.lower(),
            self.sprite_tile,
            self.sprite_palette + 1,
        )

    def has_filth(self) -> bool:
        """Returns true if there is filth attached to any edges of this tile"""
        return any(
            edge.filth_sprite_set != TileSpriteSet.NONE_0 for edge in self.edge_data
        )

    def is_dustblock(self) -> bool:
        """Returns true if the tile is a dustblock tile. This is calculated
        based on the current sprite information."""
        return SPRITE_SET_DUSTBLOCK_TILE[self.sprite_set] == self.sprite_tile

    def set_dustblock(self) -> None:
        """Update :attr:`sprite_tile` and :attr:`sprite_palette` to be the dustblock matching
        the current :attr:`sprite_set`.

        Raises:
            ValueError: If there is no dustblock tile for the current sprite set.
        """
        sprite_tile = SPRITE_SET_DUSTBLOCK_TILE[self.sprite_set]
        if sprite_tile == -1:
            raise ValueError("current sprite set does not have a dustblock tile")
        self.sprite_tile = sprite_tile
        self.sprite_palette = 1

    def transform(self, mat: TxMatrix) -> None:
        """Performs a flip and rotation as dictated by the tranformation matrix.
        Does not do any kind of scaling or skewing beyond that. Use :meth:`upscale`
        if you want to increase tile scale.

        Attributes:
            mat: The transformation matrix. See :meth:`dustmaker.level.Level.transform`
        """
        oshape = self.shape
        flipped = mat.flipped
        if flipped:
            # horizontally flip
            if self.shape == TileShape.FULL:
                self.edge_data[TileSide.LEFT], self.edge_data[TileSide.RIGHT] = (
                    self.edge_data[TileSide.RIGHT],
                    self.edge_data[TileSide.LEFT],
                )
            elif self.shape <= TileShape.SMALL_8:
                # No need to fix edge data order as SHAPE_ORDERED_SIDES uniquely
                # maps the transformed sides to original sides.
                self.shape = TileShape(1 + ((self.shape - TileShape.BIG_1) ^ 8) % 16)
            else:
                # Flipping swaps clockwise direction so we do the same.
                s1 = SHAPE_ORDERED_SIDES[self.shape][1]
                s2 = SHAPE_ORDERED_SIDES[self.shape][2]
                self.edge_data[s1], self.edge_data[s2] = (
                    self.edge_data[s2],
                    self.edge_data[s1],
                )
                self.shape = TileShape(17 + ((self.shape - TileShape.HALF_A) ^ 3))

            # flips swap clockwise to counterclockwise, reverse directed data
            for edge in self.edge_data:
                edge.caps = edge.caps[::-1]
                edge.filth_caps = edge.filth_caps[::-1]
                edge.angles = (-edge.angles[1], -edge.angles[0])
                edge.filth_angles = (-edge.filth_angles[1], -edge.filth_angles[0])

        angle = int(
            round(
                math.atan2(mat[1][1], (-1 if flipped else 1) * mat[1][0]) / math.pi * 2
            )
        )
        angle = (-angle + 1) & 0x3

        if self.shape == TileShape.FULL:
            pass
        elif self.shape <= TileShape.SMALL_4:
            self.shape = TileShape(1 + ((self.shape - TileShape.BIG_1) + angle * 2) % 8)
        elif self.shape <= TileShape.SMALL_8:
            self.shape = TileShape(9 + ((self.shape - TileShape.BIG_5) - angle * 2) % 8)
        else:
            self.shape = TileShape(17 + ((self.shape - TileShape.HALF_A) + angle) % 4)

        og_edge_data = [self.edge_data[side] for side in SHAPE_ORDERED_SIDES[oshape]]
        self.edge_data = [TileEdgeData() for _ in TileSide]

        for i, side in enumerate(SHAPE_ORDERED_SIDES[self.shape]):
            if self.shape == TileShape.FULL:
                i = (i - angle) & 3
            self.edge_data[side] = og_edge_data[i]

    def upscale(self, factor: int) -> Generator[Tuple[int, int, "Tile"], None, None]:
        """
        Upscales a tile, returning a list of (dx, dy, tile) tuples giving
        the relative position of the upscaled tiles and the new tile shape.
        This is primarily used by :meth:`dustmaker.level.Level.upscale`.

        Yields:
            A three tuple (dx, dy, ntile) where (dx, dy) are the relative position
            within the scaled up `factor` x `factor` tile square formed and
            `ntile` is the tile that belongs at that position.
        """
        if factor < 1:
            return
        if factor == 1:
            yield 0, 0, copy.deepcopy(self)
            return

        def _tuple_set(data: Tuple, ind: int, value) -> Tuple:
            return tuple((value if i == ind else x) for i, x in enumerate(data))

        def _copy_side(dx: int, dy: int, tile: Tile, side: TileSide):
            """Copies a side from self into the upscaled tile. Adjust the
            caps if the edge does not leave the upscaled tile boundary.
            """
            edge_data = copy.deepcopy(self.edge_data[side])

            cw_ind = SIDE_CLOCKWISE_INDEX[side]
            for dr in range(2):
                vert_a = SHAPE_VERTEXES[tile.shape][(cw_ind + 1 - dr) & 0x3]
                vert_b = SHAPE_VERTEXES[tile.shape][(cw_ind + dr) & 0x3]
                x = 2 * dx + vert_b[0]
                y = 2 * dy + vert_b[1]
                if vert_a[0] != vert_b[0] and x in (0, factor * 2):
                    continue
                if vert_a[1] != vert_b[1] and y in (0, factor * 2):
                    continue

                edge_data.caps = _tuple_set(edge_data.caps, dr, False)  # type: ignore
                edge_data.angles = _tuple_set(edge_data.angles, dr, 0)  # type: ignore
                edge_data.filth_caps = _tuple_set(edge_data.filth_caps, dr, False)  # type: ignore
                edge_data.filth_angles = _tuple_set(edge_data.filth_angles, dr, 0)  # type: ignore

            tile.edge_data[side] = edge_data

        if self.shape == TileShape.FULL:
            for dx in range(factor):
                for dy in range(factor):
                    tile = Tile(TileShape.FULL)
                    tile.set_sprite_tuple(self.get_sprite_tuple())

                    if dx == 0:
                        _copy_side(dx, dy, tile, TileSide.LEFT)
                    if dx + 1 == factor:
                        _copy_side(dx, dy, tile, TileSide.RIGHT)
                    if dy == 0:
                        _copy_side(dx, dy, tile, TileSide.TOP)
                    if dy + 1 == factor:
                        _copy_side(dx, dy, tile, TileSide.BOTTOM)

                    yield dx, dy, tile

        elif self.shape in (TileShape.BIG_1, TileShape.SMALL_1):
            for dx in range(factor):
                ddx = dx + (factor if self.shape == TileShape.SMALL_1 else 0)
                for dy in range(ddx // 2, factor):
                    tile = Tile(TileShape.FULL)
                    tile.set_sprite_tuple(self.get_sprite_tuple())

                    if dy == ddx // 2:
                        tile.shape = TileShape.SMALL_1 if ddx % 2 else TileShape.BIG_1
                        _copy_side(dx, dy, tile, TileSide.TOP)
                    if dx == 0:
                        _copy_side(dx, dy, tile, TileSide.LEFT)
                    if dy + 1 == factor:
                        _copy_side(dx, dy, tile, TileSide.BOTTOM)

                    yield dx, dy, tile

        elif self.shape == TileShape.HALF_A:
            for dx in range(factor):
                for dy in range(dx, factor):
                    tile = Tile(TileShape.FULL)
                    tile.set_sprite_tuple(self.get_sprite_tuple())

                    if dx == dy:
                        tile.shape = TileShape.HALF_A
                        _copy_side(dx, dy, tile, TileSide.TOP)
                    if dx == 0:
                        _copy_side(dx, dy, tile, TileSide.LEFT)
                    if dy + 1 == factor:
                        _copy_side(dx, dy, tile, TileSide.BOTTOM)

                    yield dx, dy, tile

        else:
            # Otherwise transform the tile into one of the above handled cases
            # and transform the result back.
            new_shape = self.shape

            hflip = False
            if TileShape.BIG_5 <= new_shape <= TileShape.SMALL_8:
                # horizontal flip
                hflip = True
                new_shape = TileShape(new_shape - 8)

            if TileShape.BIG_1 <= new_shape <= TileShape.SMALL_4:
                rots = (new_shape - TileShape.BIG_1) // 2
            else:  # Half tile
                rots = new_shape - TileShape.HALF_A

            # Calculate rotation matrix and inverse
            mat = TxMatrix.ROTATE[-rots % 4]
            imat = TxMatrix.ROTATE[rots % 4]

            # Apply horizontal flip
            if hflip:
                mat = mat * TxMatrix.HFLIP
                imat = TxMatrix.HFLIP * imat

            # Fix up offset so transformed positions stay in upscale square.
            mat = mat.translate(
                *(max(0, -val) for val in mat.sample(factor - 1, factor - 1))
            )
            imat = imat.translate(
                *(max(0, -val) for val in imat.sample(factor - 1, factor - 1))
            )

            # Copy tile and transform it.
            ntile = copy.deepcopy(self)
            ntile.transform(mat)  # type: ignore
            assert ntile.shape in (TileShape.HALF_A, TileShape.BIG_1, TileShape.SMALL_1)

            # For each of the new upscaled tiles inverse the transformation.
            for dx, dy, tile in ntile.upscale(factor):
                tile.transform(imat)  # type: ignore
                tx, ty = imat.sample(dx, dy)
                yield (int(tx), int(ty), tile)

    def _pack_tile_data(self) -> bytes:
        """Pack the dustmaker respresentation back into the binary representation"""
        tile_data = [0 for _ in range(12)]
        for side, edge in enumerate(self.edge_data):
            vals = [edge.solid, edge.visible, *edge.caps]
            offsets = [0 + side, 4 + side, 8 + 2 * side, 9 + 2 * side]

            if side in (TileSide.LEFT, TileSide.BOTTOM):
                # Need to swap the edge bit order to match IO order
                offsets[2], offsets[3] = offsets[3], offsets[2]

            for val, off in zip(vals, offsets):
                if val:
                    tile_data[off >> 3] |= 1 << (off & 7)

            assert -0x80 <= edge.angles[0] <= 0x7F and -0x80 <= edge.angles[1] <= 0x7F
            v0, v1 = edge.angles
            if side in (TileSide.LEFT, TileSide.BOTTOM):
                v0, v1 = v1, v0
            tile_data[2 + side * 2] = v0 & 0xFF
            tile_data[3 + side * 2] = v1 & 0xFF

        assert 0 <= self.sprite_set <= 0xF
        assert 0 <= self.sprite_tile <= 0xFF
        assert 0 <= self.sprite_palette <= 0xF

        tile_data[10] = self.sprite_set + (self.sprite_palette << 4)
        tile_data[11] = self.sprite_tile

        return bytes(tile_data)

    def _unpack_tile_data(self, tile_data: bytes) -> None:
        """Unpack tile data into the representation used by dustmaker."""
        assert len(tile_data) == 12

        # Extract the edge bits
        for side, edge in enumerate(self.edge_data):

            def test_offset(off: int) -> bool:
                return bool(tile_data[off >> 3] & (1 << (off & 7)))

            edge.solid = test_offset(0 + side)
            edge.visible = test_offset(4 + side)
            edge.caps = (test_offset(8 + 2 * side), test_offset(9 + 2 * side))
            if side in (TileSide.LEFT, TileSide.BOTTOM):
                edge.caps = (edge.caps[1], edge.caps[0])

            v0, v1 = tile_data[2 + side * 2], tile_data[3 + side * 2]
            if v0 >= 0x80:
                v0 -= 0x100
            if v1 >= 0x80:
                v1 -= 0x100
            if side in (TileSide.LEFT, TileSide.BOTTOM):
                v0, v1 = v1, v0
            edge.angles = (v0, v1)

        self.sprite_set = TileSpriteSet(tile_data[10] & 0xF)
        self.sprite_tile = tile_data[11]
        self.sprite_palette = tile_data[10] >> 4

    def _pack_dust_data(self) -> bytes:
        """Pack the dustmaker respresentation back into the binary representation"""
        dust_data = [0 for _ in range(12)]

        for side, edge in enumerate(self.edge_data):
            sset, spike, caps, angles = (
                edge.filth_sprite_set,
                edge.filth_spike,
                edge.filth_caps,
                edge.filth_angles,
            )
            assert 0 <= sset <= 0xF
            assert -0x80 <= angles[0] <= 0x7F
            assert -0x80 <= angles[1] <= 0x7F

            # Normalize for Dustforce binary format
            if side in (TileSide.LEFT, TileSide.BOTTOM):
                caps = caps[::-1]
                angles = angles[::-1]

            off = 4 * side
            dust_data[off >> 3] |= (sset | (0x8 if spike else 0)) << (off & 0x7)
            dust_data[2 + side * 2] = angles[0] & 0xFF
            dust_data[3 + side * 2] = angles[1] & 0xFF
            if caps[0]:
                dust_data[10] |= 1 << (2 * side)
            if caps[1]:
                dust_data[10] |= 2 << (2 * side)

        return bytes(dust_data)

    def _unpack_dust_data(self, dust_data: bytes) -> None:
        """Unpack dust data into the representation used by dustmaker."""
        assert len(dust_data) == 12

        for side, edge in enumerate(self.edge_data):
            off = 4 * side
            val = dust_data[off >> 3] >> (off & 0x7)
            edge.filth_sprite_set = TileSpriteSet(val & 0x7)
            edge.filth_spike = bool(val & 0x8)
            edge.filth_caps = (
                bool((dust_data[10] >> (2 * side)) & 0x1),
                bool((dust_data[10] >> (2 * side)) & 0x2),
            )

            v0, v1 = dust_data[2 + side * 2], dust_data[3 + side * 2]
            if v0 >= 0x80:
                v0 -= 0x100
            if v1 >= 0x80:
                v1 -= 0x100
            edge.filth_angles = (v0, v1)

            # Normalize for Dustforce binary format
            if side in (TileSide.LEFT, TileSide.BOTTOM):
                edge.filth_caps = edge.filth_caps[::-1]
                edge.filth_angles = edge.filth_angles[::-1]


#: Mapping of :class:`TileSpriteSet` to the corresponding dustblock index
#: for that sprite set. Gives -1 if no dustblock tile is available for the
#: given sprite set.
#:
#: Type:
#:     tuple mapping :class:`TileSpriteSet` -> int
#:
#: :meta hide-value:
SPRITE_SET_DUSTBLOCK_TILE = (
    -1,  # NONE_0
    21,  # MANSION
    13,  # FOREST
    6,  # CITY
    9,  # LABORATORY
    2,  # TUTORIAL
    -1,  # NEXUS
    -1,  # NONE_7
)


#: A mapping of :class:`TileShape` to a sequence of sides.
#:
#: For: :attr:`TileShape.FULL` this is ordered clockwise starting with
#: the top side.
#:
#: For half tiles and small slants the ordering starts on the diagonal
#: edge and procedes clockwise around the tile.
#:
#: For big slants the ordering starts on the diagonal, then the opposite
#: side, then the flat side that's not present on the small slants. For
#: BIG_1 through BIG_4 this is clockwise, for BIG_5 through BIG_8 this
#: is counter-clockwise.
#:
#: Type:
#:     tuple mapping :class:`TileShape` -> (TileSide, ...)
#:
#: :meta hide-value:
SHAPE_ORDERED_SIDES = tuple(
    tuple(TileSide(x) for x in y)
    for y in (
        (0, 3, 1, 2),
        (0, 1, 2),
        (0, 1),
        (3, 2, 0),
        (3, 2),
        (1, 0, 3),
        (1, 0),
        (2, 3, 1),
        (2, 3),
        (0, 1, 3),
        (0, 1),
        (2, 3, 0),
        (2, 3),
        (1, 0, 2),
        (1, 0),
        (3, 2, 1),
        (3, 2),
        (0, 1, 2),
        (1, 2, 0),
        (1, 0, 3),
        (0, 3, 1),
    )
)


#: Mapping of :class:`TileShape` to the vertex coordinates of the tile in
#: half-tile units. Vertexes are listed top-left, top-right, bottom-right,
#: and bottom-left order.
#:
#: Type:
#:     tuple mapping :class:`TileShape` -> ((int, int), (int, int), (int, int), (int, int))
#:
#: :meta hide-value:
SHAPE_VERTEXES = (
    ((0, 0), (2, 0), (2, 2), (0, 2)),  # FULL
    # Slants are rotations of each other
    ((0, 0), (2, 1), (2, 2), (0, 2)),  # BIG_1
    ((0, 1), (2, 2), (2, 2), (0, 2)),  # SMALL_1
    ((0, 0), (2, 0), (1, 2), (0, 2)),  # BIG_2
    ((0, 0), (1, 0), (0, 2), (0, 2)),  # SMALL_2
    ((0, 0), (2, 0), (2, 2), (0, 1)),  # BIG_3
    ((0, 0), (2, 0), (2, 1), (0, 0)),  # SMALL_3
    ((1, 0), (2, 0), (2, 2), (0, 2)),  # BIG_4
    ((2, 0), (2, 0), (2, 2), (1, 2)),  # SMALL_4
    ((0, 1), (2, 0), (2, 2), (0, 2)),  # BIG_5
    ((0, 2), (2, 1), (2, 2), (0, 2)),  # SMALL_5
    ((0, 0), (2, 0), (2, 2), (1, 2)),  # BIG_6
    ((1, 0), (2, 0), (2, 2), (2, 2)),  # SMALL_6
    ((0, 0), (2, 0), (2, 1), (0, 2)),  # BIG_7
    ((0, 0), (2, 0), (2, 0), (0, 1)),  # SMALL_7
    ((0, 0), (1, 0), (2, 2), (0, 2)),  # BIG_8
    ((0, 0), (0, 0), (1, 2), (0, 2)),  # SMALL_8
    # Slopes are special because they always have a top and bottom so they
    # don't simply rotate between each one. The repeated coordinate is
    # always placed so it makes a left or right edge null.
    ((0, 0), (2, 2), (2, 2), (0, 2)),  # HALF_A
    ((0, 0), (2, 0), (2, 0), (0, 2)),  # HALF_B
    ((0, 0), (2, 0), (2, 2), (0, 0)),  # HALF_C
    ((0, 2), (2, 0), (2, 2), (0, 2)),  # HALF_D
)


#: Mapping of :class:`TileSide` to the index of that tile side when sides are
#: listed in clockwise order. This is useful for computing the edge vertexes
#: for a given side from :attr:`SHAPE_VERTEXES`.
#:
#: Type:
#:     tuple mapping :class:`TileSide` -> int
#:
#: Examples:
#:     ::
#:
#:         shape, side = TileShape.BIG_1, TileSide.TOP
#:         ind = SIDE_CLOCKWISE_INDEX[side]
#:         vert_a = SHAPE_VERTEXES[shape][ind]
#:         vert_b = SHAPE_VERTEXES[shape][(ind + 1) % 4]
#:         # vert_a = (0, 0), vert_b = (2, 1)
#:
#: :meta hide-value:
SIDE_CLOCKWISE_INDEX = (
    0,  # TOP
    2,  # BOTTOM
    3,  # LEFT
    1,  # RIGHT
)



""" Module providing methods for reading Dustforce binary formats including
level files.
"""
import io
from typing import Any, Dict, List, Tuple
import zlib

from .bitio import BitIOReader
from .entity import Entity
from .level import Level, LevelType
from .exceptions import LevelParseException
from .prop import Prop
from .tile import Tile, TileShape
from .variable import (
    Variable,
    VariableArray,
    VariableBool,
    VariableFloat,
    VariableInt,
    VariableString,
    VariableStruct,
    VariableType,
    VariableUInt,
    VariableVec2,
)


class DFReader(BitIOReader):
    """Helper class to read Dustforce binary files"""

    def read_expect(self, data: bytes) -> None:
        """Ensure the next bytes match `data`

        Raises:
            LevelParseException: If the read bytes do not match `data`.
        """
        if data != self.read_bytes(len(data)):
            raise LevelParseException("unexpected data")

    def read_float(self, ibits: int, fbits: int) -> float:
        """Read a float in the Dustforce format

        Arguments:
            ibits (int): Number of integer bits
            fbits (int): Number of fractional bits
        """
        sign = 1 - 2 * self.read(1)
        ipart = self.read(ibits - 1)
        fpart = self.read(fbits)
        return sign * ipart + 1.0 * fpart / (1 << (fbits - 1))

    def read_6bit_str(self) -> str:
        """Read a '6-bit' string. These are strings with length between
        0 and 63 inclusive that contain only alpha-numeric lower and uppercase
        characters in addition to '_' and '{'.
        """
        ln = self.read(6)
        chrs = []
        for _ in range(ln):
            v = self.read(6)
            if v < 10:
                chrs.append(chr(ord("0") + v))
            elif v < 36:
                chrs.append(chr(ord("A") + v - 10))
            elif v == 36:
                chrs.append("_")
            elif v < 63:
                chrs.append(chr(ord("a") + v - 37))
            else:
                chrs.append("{")
        return "".join(chrs)

    def read_variable(self, vtype: VariableType) -> Variable:
        """Read a variable of a given type.

        Arguments:
            vtype (VariableType): The type of variable to read
        """
        max_width = (2 ** 16) - 1
        if vtype == VariableType.NULL:
            raise LevelParseException("unexpected null variable")
        if vtype == VariableType.BOOL:
            return VariableBool(self.read(1) == 1)
        if vtype == VariableType.UINT:
            return VariableUInt(self.read(32))
        if vtype == VariableType.INT:
            return VariableInt(self.read(32, True))
        if vtype == VariableType.FLOAT:
            return VariableFloat(self.read_float(32, 32))

        if vtype == VariableType.STRING:
            slen = self.read(16)
            return VariableString(self.read_bytes(slen))

        if vtype == VariableType.VEC2:
            f0 = self.read_float(32, 32)
            f1 = self.read_float(32, 32)
            return VariableVec2((f0, f1))

        if vtype == VariableType.ARRAY:
            atype = VariableType(self.read(4))
            alen = self.read(16)
            val: List[Variable] = []

            if atype != VariableType.STRING:
                val.extend(self.read_variable(atype) for _ in range(alen))
            else:
                while alen > 0:
                    var_value_ctns = []
                    while alen > 0:
                        alen -= 1
                        value_part = self.read_variable(atype).value
                        var_value_ctns.append(value_part)
                        if len(value_part) < max_width:
                            break

                    val.append(VariableString(b"".join(var_value_ctns)))

            return VariableArray(Variable._TYPES[atype], val)

        if vtype == VariableType.STRUCT:
            result: Dict[str, Variable] = {}
            while True:
                vtype = VariableType(self.read(4))
                if vtype == VariableType.NULL:
                    break

                var_name = self.read_6bit_str()
                if vtype != VariableType.STRING:
                    result[var_name] = self.read_variable(vtype)
                    continue

                var_value_ctns = []
                while True:
                    value_part = self.read_variable(vtype)
                    var_value_ctns.append(value_part.value)
                    if len(value_part.value) < max_width:
                        break

                    self.read(4)
                    self.read_6bit_str()

                result[var_name] = VariableString(b"".join(var_value_ctns))

            return VariableStruct(result)

        raise LevelParseException("unknown var type")

    def read_variable_map(self) -> Dict[str, Variable]:
        """Convenience method equivalent to `read_variable(VariableType.STRUCT).value`"""
        return self.read_variable(VariableType.STRUCT).value

    def read_segment(self, level: Level, xoffset: int, yoffset: int) -> None:
        """Read segment data into the passed level. In most cases you should
        just use :meth:`read_level` instead of this method.

        Arguments:
            level (Level): The level object to read data into
            xoffset (int): The segment x-offset in tiles
            yoffset (int): The segment y-offset in tiles
        """
        start_index = self.bit_tell()

        segment_size = self.read(32)
        version = self.read(16)
        xoffset += self.read(8) * 16
        yoffset += self.read(8) * 16
        segment_width = self.read(8)  # pylint: disable=unused-variable

        if version > 4:
            level_uid = self.read(32)  # pylint: disable=unused-variable
            dust_filth = self.read(16)  # pylint: disable=unused-variable
            enemy_filth = self.read(16)  # pylint: disable=unused-variable
        if version > 5:
            tile_surface = self.read(16)  # pylint: disable=unused-variable
            dustblock_filth = self.read(16)  # pylint: disable=unused-variable

        flags = self.read(32)
        if flags & 1:
            layers = self.read(8)
            for _ in range(layers):
                layer = self.read(8)
                tiles = self.read(10)

                for _ in range(tiles):
                    txpos = self.read(5)
                    typos = self.read(5)
                    shape = self.read(5)
                    tile_flags = self.read(3)
                    data = self.read_bytes(12)
                    level.tiles[(layer, xoffset + txpos, yoffset + typos)] = Tile(
                        TileShape(shape & 0x1F),
                        tile_flags=tile_flags,
                        _tile_data=data,
                    )

        if flags & 2:
            dusts = self.read(10)
            for _ in range(dusts):
                txpos = self.read(5)
                typos = self.read(5)
                data = self.read_bytes(12)
                tile = level.tiles.get((19, xoffset + txpos, yoffset + typos))
                if tile is not None:
                    tile._unpack_dust_data(data)

        if flags & 8:
            props = self.read(16)
            for _ in range(props):
                id_num = self.read(32, True)
                if id_num < 0:
                    continue

                layer = self.read(8)
                layer_sub = self.read(8)

                scale = 1.0
                if version > 6 or level.level_type == LevelType.DUSTMOD:
                    x_sgn = self.read(1)
                    x_int = self.read(27)
                    x_scale = (self.read(4) & 0x7) ^ 0x4
                    y_sgn = self.read(1)
                    y_int = self.read(27)
                    y_scale = (self.read(4) & 0x7) ^ 0x4

                    xpos = (-1.0 if x_sgn != 0 else 1.0) * x_int
                    ypos = (-1.0 if y_sgn != 0 else 1.0) * y_int

                    scale_lg = x_scale * 7 + y_scale
                    scale = pow(50.0, (scale_lg - 32.0) / 24.0)
                else:
                    xpos = self.read_float(28, 4)
                    ypos = self.read_float(28, 4)

                rotation = self.read(16)
                flip_x = self.read(1) != 0
                flip_y = self.read(1) != 0
                prop_set = self.read(8)
                prop_group = self.read(12)
                prop_index = self.read(12)
                palette = self.read(8)

                # Default DF behavior is to overwrite repeated props
                level.props.pop(id_num, None)
                level.add_prop(
                    layer,
                    xpos,
                    ypos,
                    Prop(
                        layer_sub,
                        rotation,
                        flip_x,
                        flip_y,
                        scale,
                        prop_set,
                        prop_group,
                        prop_index,
                        palette,
                    ),
                    id_num=id_num,
                )

        if flags & 4:
            entities = []
            num_entities = self.read(16)
            has_extended_names = False
            for _ in range(num_entities):
                id_num = self.read(32, True)
                if id_num < 0:
                    continue

                etype = self.read_6bit_str()
                if etype == "entity" and version > 7:
                    has_extended_names = True
                xpos = self.read_float(32, 8)
                ypos = self.read_float(32, 8)
                rotation = self.read(16)
                layer = self.read(8)
                faceX = self.read(1)
                faceY = self.read(1)
                visible = self.read(1)
                variables = self.read_variable_map()

                entities.append(
                    (
                        etype,
                        xpos,
                        ypos,
                        (
                            variables,
                            rotation,
                            layer,
                            faceX == 0,
                            faceY == 0,
                            visible == 1,
                        ),
                        id_num,
                    )
                )

            # Read in extended names if present and add entities to level
            if has_extended_names:
                self.bit_seek(start_index + (segment_size - 4) * 8)
                extra_names_index = self.read(32)
                self.skip(-extra_names_index - 32)

            for etype, xpos, ypos, eargs, id_num in entities:
                if has_extended_names and etype == "entity":
                    etype = self.read_6bit_str()
                level.add_entity(
                    xpos,
                    ypos,
                    Entity._from_raw(etype, *eargs),
                    id_num,
                )

        self.bit_seek(start_index + segment_size * 8)

    def read_region(self, level: Level) -> None:
        """Read region data into the passed level. In most cases you should
        just use :meth:`read_level` instead of this method.

        Arguments:
            level (Level): The level object to read data into
        """
        region_len = self.read(32)
        uncompressed_len = self.read(32)  # pylint: disable=unused-variable
        offx = self.read(16, True)
        offy = self.read(16, True)
        version = self.read(16)  # pylint: disable=unused-variable
        segments = self.read(16)
        has_backdrop = self.read(8) != 0

        sub_reader = DFReader(
            io.BytesIO(zlib.decompress(self.read_bytes(region_len - 17)))
        )
        for _ in range(segments):
            sub_reader.align()
            sub_reader.read_segment(level, offx * 256, offy * 256)

        if has_backdrop:
            if level.backdrop is None:
                raise ValueError("no backdrop present in level")
            sub_reader.align()
            sub_reader.read_segment(level.backdrop, offx * 16, offy * 16)

    def _read_metadata(self) -> Dict[str, Any]:
        """Read a metadata block (DF_MTD) into a JSON payload."""
        self.read_expect(b"DF_MTD")

        version = self.read(16)
        region_offset = self.read(32)  # pylint: disable=unused-variable
        entityUid = self.read(32)
        propUid = self.read(32)
        deprecatedSaveUid = self.read(32)
        regionUidDeprecated = self.read(32)
        return {
            "version": version,
            "entityUid": entityUid,
            "propUid": propUid,
            "deprecatedSaveUid": deprecatedSaveUid,
            "regionUidDeprecated": regionUidDeprecated,
        }

    def read_var_file(self, header: bytes) -> Dict[str, Variable]:
        """Reads a variable mapping with a given header. There are several
        file types that Dustforce use that are expressed this way including
        notably "stats1" (header=b"DF_STA") and "config" (header=b"DF_CFG").

        Arguments:
            header (bytes): The expected file header at the start of the stream.
                Just pass b"" if you've already read and checked the header.
        """
        self.read_expect(header)
        version = self.read(16)  # pylint: disable=unused-variable
        statSize = self.read(32)  # pylint: disable=unused-variable
        return self.read_variable_map()

    def read_level_ex(self) -> Tuple[Level, List[int]]:
        """Extended version of :meth:`read_level`.

        Read level file metadata into a :class:`Level` object while
        extracting additional metadata so that the rest of the data can be
        ingested in an opaque way. `read_level_ex` ends with the reader
        byte-aligned. The entirety of the region data can be read subsequently
        with reader.read_bytes(region_bytes) or using :meth:`read_region`.

        This can be used with :meth:`dustmaker.dfwriter.DFWriter.write_level_ex` to modify
        level metadata without reading in region data.

        Example: ::

            # Re-write level metadata without reading in region data.
            level, region_offsets = reader.read_level_ex()
            region_data = reader.read_bytes(region_offsets[-1])
            ...
            writer.write_level_ex(level, region_offsets, region_data)

        Example: ::

            # Manually read region data
            level, region_offsets = reader.read_level_ex()
            for _ in region_offsets[:-1]:
                reader.read_region(level)

        Returns:
            (level, region_offsets) tuple

            - level
                :class:`dustmaker.level.Level` object with metadata (e.g.
                level.variables and level.sshot) filled in.
            - region_offsets
                list of byte offsets of each region from the current stream position
                (which is aligned). The last element of this array is the end of the
                region data and does not correspond to a region itself.
        """
        assert self.aligned()
        start_index = self.bit_tell()
        self.read_expect(b"DF_LVL")

        version = self.read(16)
        if version <= 42:
            raise LevelParseException("unsupported level version")

        filesize = self.read(32)  # pylint: disable=unused-variable
        num_regions = self.read(32)
        meta = self._read_metadata()  # pylint: disable=unused-variable

        sshot_data = b""
        if version > 43:
            sshot_len = self.read(32)
            sshot_data = self.read_bytes(sshot_len)

        level = Level()
        level._next_id = meta["entityUid"]
        level.variables = self.read_variable_map()
        level.sshot = sshot_data

        region_offsets = [self.read(32) for _ in range(num_regions)]
        self.align()

        region_offsets.append(filesize - ((self.bit_tell() - start_index) >> 3))
        return level, region_offsets

    def read_level(self, *, metadata_only: bool = False) -> Level:
        """Read a level data stream and return the :class:`dustmaker.level.Level`
        object.

        Arguments:
            metadata_only (bool, optional): If set to True only the variables
                and sshot data will be set in the returned :class:`Level`.

        Raises:
            LevelParseException: Parser ran into unexpected data.
        """
        level, region_offsets = self.read_level_ex()
        if metadata_only:
            return level

        for _ in range(len(region_offsets) - 1):
            self.read_region(level)
        return level


def read_level(data: bytes) -> Level:
    """Convenience function to read in a level from bytes directly

    Arguments:
        data (bytes): The data source for the level

    Returns:
        The parsed Level object.
    """
    with DFReader(io.BytesIO(data), noclose=True) as reader:
        return reader.read_level()



#!/usr/bin/env python3
"""
Sample script to perform level transforms
"""
import argparse
import sys

from dustmaker import DFReader, DFWriter
from dustmaker.transform import TxMatrix
from dustmaker.cmd.common import (
    run_utility,
    CliUtility,
)
from dustmaker.level import LevelType


class Transform(CliUtility):
    """CLI utility for peforming affine transformations on levels"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        """Read CLI arguments"""
        parser.description = (
            "upscales, rotates, flips, and then shifts a level (in that order)."
        )
        parser.add_argument("input_level")
        parser.add_argument("output_level")
        parser.add_argument(
            "--upscale",
            type=int,
            default=1,
            required=False,
            help="how much to upscale level size",
        )
        parser.add_argument(
            "--rotate",
            type=int,
            default=0,
            required=False,
            help="how many 90 degree clockwise rotations to perform (use negative for counterclockwise)",
        )
        parser.add_argument(
            "--vflip",
            action="store_const",
            const=True,
            default=False,
            required=False,
            help="vertically flip result",
        )
        parser.add_argument(
            "--hflip",
            action="store_const",
            const=True,
            default=False,
            required=False,
            help="horizontally filp result",
        )
        parser.add_argument(
            "--translate-x",
            type=float,
            default=0,
            required=False,
            help="how much to shift the level elements' x-coordinates in pixels",
        )
        parser.add_argument(
            "--translate-y",
            type=float,
            default=0,
            required=False,
            help="how much to shift the level elements' y-coordinates in pixels",
        )

    def main(self, args):
        """Transform CLI entrypoint"""
        with DFReader(open(args.input_level, "rb")) as reader:
            level = reader.read_level()

        # Start out with the rotation matrix
        mat = TxMatrix.ROTATE[args.rotate % 4]

        # Perform the flips
        if args.vflip:
            mat = TxMatrix.VFLIP * mat

        if args.hflip:
            mat = TxMatrix.HFLIP * mat

        # Add in the translations
        mat = mat.translate(args.translate_x, args.translate_y)

        if args.upscale > 1:
            # Need dustmod level type for proper scaling of entities/props
            if level.level_type == LevelType.NORMAL:
                level.level_type = LevelType.DUSTMOD
            level.upscale(args.upscale, mat=mat)
        else:
            level.transform(mat)

        with DFWriter(open(args.output_level, "wb")) as writer:
            writer.write_level(level)


if __name__ == "__main__":
    sys.exit(run_utility(Transform))



""" Module defining shared exception classes. """


class LevelException(Exception):
    """Top level dustmaker exception."""


class LevelParseException(LevelException):
    """Exception indicating an error reading a level file."""



"""
Module defining basic entity representations as well as a custom entity
object for each entity in Dustforce.
"""
import copy
from enum import IntEnum
import math
from typing import cast, Dict, Optional, Tuple, Type, TypeVar

from .transform import TxMatrix
from .variable import (
    Variable,
    VariableArray,
    VariableBool,
    VariableFloat,
    VariableInt,
    VariableString,
    VariableStruct,
    VariableUInt,
    VariableVec2,
)

T = TypeVar("T")


def _get_python_prop_type(prop_type: Type[Variable]) -> str:
    """Returns the python type name for variable types for use
    with documentation."""
    if prop_type is VariableBool:
        return "bool"
    if prop_type in (VariableInt, VariableUInt):
        return "int"
    if prop_type is VariableFloat:
        return "float"
    if prop_type is VariableString:
        return "bytes"
    if prop_type is VariableVec2:
        return "(float, float)"
    if prop_type is VariableStruct:
        return "dict[str, Variable]"
    if prop_type is VariableArray:
        return "MutableSequence"
    raise TypeError("unexpected variable type")


def bind_prop(
    prop_name: str,
    prop_type: Type[Variable],
    default: T,
    doc: Optional[str] = None,
    doc_add_type=True,
    objtype=False,
) -> property:
    """Define getters and setters for a named property backed by the map's
    var mapping.

    :meta private:
    """

    # Removed "-> T" type annotation on this because it was making docs ugly.
    def _get(self):
        """get the property with given default"""
        var = self.variables.get(prop_name)
        if var is None:
            result = default() if callable(default) else default
            if objtype:
                self.variables[prop_name] = result
            return result
        if not isinstance(var, prop_type):
            raise ValueError("unexpected property type")
        if objtype:
            return cast(T, var)
        return cast(T, var.value)

    def _set(self, val: T) -> None:
        """set the property"""
        self.variables[prop_name] = val if objtype else prop_type(val)

    def _del(self) -> None:
        """delete the property. Ignored if the property is not set."""
        self.variables.pop(prop_name, None)

    if doc is None:
        doc = f"Wrapper around `variables['{prop_name}']` of type :class:`{prop_type.__name__}`."
    if doc_add_type:
        doc = f"{_get_python_prop_type(prop_type)}: {doc}"
    if not objtype:
        doc = f"{doc} Defaults to `{default}`."

    return property(fget=_get, fset=_set, fdel=_del, doc=doc)


def bind_prop_arr(
    prop_name: str,
    elem_type: Type[Variable],
    doc: Optional[str] = None,
    doc_add_type=True,
) -> property:
    """Convenience wrapper around bind_prop for array properties

    :meta private:
    """
    if doc is None:
        doc = f"Wrapper around `variables['{prop_name}']` of type `VariableArray[{elem_type.__name__}]`."
    if doc_add_type:
        doc = f"MutableSequence[{_get_python_prop_type(elem_type)}]: {doc}"

    return bind_prop(
        prop_name,
        VariableArray,
        lambda: VariableArray(elem_type),
        doc=doc,
        doc_add_type=False,
        objtype=True,
    )


class Entity:
    """Base class representing an entity object in a map. Commonly used
    entities have subclasses of :class:`Entity` to represent them. Those that
    do not will simply be of type :class:`Entity` and have their :attr:`etype`
    field set to control how the Dustforce engine will interpret the entity.

    To add new Entity types simply extend this class and include a `TYPE_IDENTIFIER`
    class attribute that matches the type identifier used internally by Dustforce.
    The level reader and writer will then automatically use these types (as long as
    the classes have been initialized). Additionally you can override the
    :meth:`remap_ids` and :meth:`transform` methods to handle any special processing
    this entity type requires. If you make these changes consider contributing your
    entity specialization as a pull request.

    Attributes:
        etype (str): The entity type name. This typically matches the `TYPE_IDENTIFIER`
            attribute of the concrete type of this object.
        variables (dict[str, Variable]): Persist data variable mapping for this entity
        rotation (16-bit uint): Clockwise rotation of the entity ranging from 0 to 0xFFFF.
            0x4000 corresponds to a 90 degree rotation, 0x8000 to 180 degrees, 0xC000
            to 270 degrees. This rotation is logically applied after any flips have
            been applied.
        layer (8-bit uint): Layer to render the entity in
        flip_x (bool): Flip the entity horizontally
        flip_y (bool): Flip the entity vertically
        visible (bool): Is the entity visible
    """

    _known_types: Dict[str, Type["Entity"]] = {}

    @classmethod
    def __init_subclass__(cls) -> None:
        """
        Register subclasses as a known type.
        """
        etype = getattr(cls, "TYPE_IDENTIFIER", None)
        if etype is not None:
            cls._known_types[etype] = cls

    @classmethod
    def _from_raw(
        cls,
        etype: str,
        variables: Dict[str, Variable],
        rotation: int,
        layer: int,
        flip_x: bool,
        flip_y: bool,
        visible: bool,
    ) -> "Entity":
        """Construct an entity from its map format definition."""
        subcls = cls._known_types.get(etype)
        if subcls is not None:
            return subcls(variables, rotation, layer, flip_x, flip_y, visible)
        entity = Entity(variables, rotation, layer, flip_x, flip_y, visible)
        entity.etype = etype
        return entity

    def __init__(
        self,
        variables: Optional[Dict[str, Variable]] = None,
        rotation=0,
        layer=18,
        flip_x=False,
        flip_y=False,
        visible=True,
    ) -> None:
        self.etype = getattr(self, "TYPE_IDENTIFIER", "unknown_type")
        self.variables = copy.deepcopy(variables) if variables is not None else {}
        self.rotation = rotation
        self.layer = layer
        self.flip_x = flip_x
        self.flip_y = flip_y
        self.visible = visible

    def __repr__(self) -> str:
        return "Entity: (%s, %d, %d, %d, %d, %d, %s)" % (
            self.etype,
            self.rotation,
            self.layer,
            self.flip_x,
            self.flip_y,
            self.visible,
            repr(self.variables),
        )

    def remap_ids(self, id_map: Dict[int, int]) -> None:
        """Overridable method to allow an entity to remap any internally stored
        IDs."""

    def transform(self, mat: TxMatrix) -> None:
        """Generic transform implementation for entities. Transforms the Entity
        :attr:`rotation` and :attr:`flip_y` attributes (:attr:`flip_x` is redundant).

        Many subtypes will perform additional transformations on their :attr:`variables`.
        """
        self.rotation = self.rotation - int(0x10000 * mat.angle / math.pi / 2) & 0xFFFF
        if mat.flipped:
            self.flip_y = not self.flip_y
            self.rotation = -self.rotation & 0xFFFF


class Emitter(Entity):
    """Emitter entity class"""

    TYPE_IDENTIFIER = "entity_emitter"

    e_rotation = bind_prop("e_rotation", VariableInt, 0)
    draw_depth_sub = bind_prop("draw_depth_sub", VariableUInt, 0)
    r_rotation = bind_prop("r_rotation", VariableBool, False)
    r_area = bind_prop("r_area", VariableBool, False)
    width = bind_prop("width", VariableInt, 480)
    height = bind_prop("height", VariableInt, 480)
    emitter_id = bind_prop("emitter_id", VariableUInt, 0)


class CheckPoint(Entity):
    """Checkpoint entity class"""

    TYPE_IDENTIFIER = "check_point"

    def transform(self, mat: TxMatrix) -> None:
        """Transform the trigger area"""
        super().transform(mat)

        areas = self.trigger_areas
        for i, pos in enumerate(areas):
            # trigger_area locations are relative the entity itself so should
            # not be sampled with offset.
            areas[i] = mat.sample(*pos, with_offset=False)

    trigger_areas = bind_prop_arr("trigger_area", VariableVec2)


class EndZone(CheckPoint):
    """Proximity based end zone (purple flag) entity class"""

    TYPE_IDENTIFIER = "level_end_prox"

    finished = bind_prop("finished", VariableBool, False)


class Trigger(Entity):
    """Trigger entity entity class"""

    TYPE_IDENTIFIER = "base_trigger"

    width = bind_prop("width", VariableInt, 500)


class FogTrigger(Trigger):
    """Fog trigger entity class"""

    TYPE_IDENTIFIER = "fog_trigger"

    def _fog_by_index(
        self, index: int, val: Optional[Tuple[int, float]]
    ) -> Tuple[int, float]:
        """Internal method to get the fog values at a given index corresponding
        to either a layer or sublayer"""
        if index < 0 or index >= 26 * 21:
            raise IndexError("invalid fog index")

        colours = self.variables.setdefault("fog_trigger", VariableArray(VariableUInt))
        if not isinstance(colours, VariableArray):
            raise ValueError("fog_trigger variable not an array")
        pers = self.variables.setdefault("fog_per", VariableArray(VariableFloat))
        if not isinstance(pers, VariableArray):
            raise ValueError("fog_per variable not an array")

        while index >= len(colours):
            colours.append(0x111118)
        while index >= len(pers):
            pers.append(0.0)

        result = (colours[index], pers[index])
        if val is not None:
            colours[index], pers[index] = val
        return result

    def normalize(self) -> None:
        """Resizes :attr:`colours` and :attr:`pers` arrays to be the
        correct length for the given :attr:`has_sub_layers` setting.
        """
        want_length = 21 * 26 if self.has_sub_layers else 21
        while len(self.colours) > want_length:
            self.colours.pop()
        while len(self.pers) > want_length:
            self.pers.pop()
        while len(self.colours) < want_length:
            self.colours.append(0x111118)
        while len(self.pers) < want_length:
            self.pers.append(0.0)

    @staticmethod
    def get_layer_index(layer: int, sublayer: Optional[int] = None) -> int:
        """Helper function to find fog data for a given layer/sublayer.

        Arguments:
            layer (int): Must be between 0 and 20 inclusive.
            sublayer (int): Must be between 0 and 24 inclusive.

        Returns:
            Index into :attr:`colours` and :attr:`pers` where fog data is
            stored for the given `layer` and (if present) `sublayer`.
        """
        assert 0 <= layer <= 20
        if sublayer is None:
            return layer
        assert 0 <= sublayer <= 20
        return (sublayer + 1) * 21 + layer

    speed = bind_prop("speed", VariableFloat, 5.0)
    gradient = bind_prop_arr("gradient", VariableUInt)
    gradient_middle = bind_prop("gradient_middle", VariableFloat, 0.0)
    star_bottom = bind_prop("star_bottom", VariableFloat, 0.0)
    star_middle = bind_prop("star_middle", VariableFloat, 0.4)
    star_top = bind_prop("star_top", VariableFloat, 1.0)
    has_sub_layers = bind_prop(
        "has_sub_layers",
        VariableBool,
        False,
        "Controls if sublayer fog data is enabled for this trigger",
    )
    colours = bind_prop_arr(
        "fog_trigger",
        VariableUInt,
        "Fog colour in 0xRRGGBB format for each (sub)layer.",
    )
    pers = bind_prop_arr(
        "fog_per",
        VariableFloat,
        "Mixing coefficient for the fog each (sub)layer from 0.0 to 1.0.",
    )


class AmbienceTrigger(Trigger):
    """Ambience trigger entity class"""

    TYPE_IDENTIFIER = "ambience_trigger"

    speed = bind_prop("ambience_speed", VariableFloat, 5)
    sound_names = bind_prop_arr("sound_ambience_names", VariableString)
    sound_vols = bind_prop_arr("sound_ambience_vol", VariableFloat)


class MusicTrigger(Trigger):
    """Music trigger entity class"""

    TYPE_IDENTIFIER = "music_trigger"

    speed = bind_prop("music_speed", VariableFloat, 5)
    sound_names = bind_prop_arr("sound_music_names", VariableString)
    sound_vols = bind_prop_arr("sound_music_vol", VariableFloat)


class SpecialTrigger(Trigger):
    """Max special trigger entity class"""

    TYPE_IDENTIFIER = "special_trigger"


class TextTrigger(Entity):
    """Text trigger entity class"""

    TYPE_IDENTIFIER = "text_trigger"

    hide = bind_prop("hide", VariableBool, False)
    text = bind_prop("text_string", VariableString, b"")


class DeathZone(Entity):
    """Death zone entity class"""

    TYPE_IDENTIFIER = "kill_box"

    def transform(self, mat: TxMatrix) -> None:
        """Transform the death zone width and height"""
        super().transform(mat)
        tw, th = mat.sample(self.width, self.height, with_offset=False)
        self.width = int(abs(tw))
        self.height = int(abs(th))

    width = bind_prop("width", VariableInt, 0)
    height = bind_prop("height", VariableInt, 0)


class AIController(Entity):
    """AI controller node entity class"""

    TYPE_IDENTIFIER = "AI_controller"

    def remap_ids(self, id_map: Dict[int, int]) -> None:
        """Remap the puppet id."""
        super().remap_ids(id_map)
        self.puppet = id_map.get(self.puppet, 0)

    def transform(self, mat: TxMatrix) -> None:
        """Transform the controller waypoints."""
        super().transform(mat)

        nodes = self.nodes
        for i, pos in enumerate(nodes):
            nodes[i] = mat.sample(*pos)

    nodes = bind_prop_arr("nodes", VariableVec2)
    node_wait_times = bind_prop_arr("nodes_wait_time", VariableInt)
    puppet = bind_prop("puppet_id", VariableUInt, 0)


class CameraNodeType(IntEnum):
    """
    Enum defining the different camera node types
    """

    NORMAL = 1
    DETACH = 2
    CONNECT = 3
    INTEREST = 4
    FORCE_CONNECT = 5


class CameraNode(Entity):
    """Camera node entity class"""

    TYPE_IDENTIFIER = "camera_node"

    def remap_ids(self, id_map: Dict[int, int]) -> None:
        """Remap the connected camera node IDs."""
        super().remap_ids(id_map)
        nds = self.nodes
        for i, nd_id in enumerate(nds):
            nds[i] = id_map.get(nd_id, 0)

    def transform(self, mat: TxMatrix) -> None:
        """Transform the camera zoom and width"""
        super().transform(mat)
        scale = mat.scale
        self.zoom = int(round(self.zoom * scale))
        self.width = int(round(self.width * scale))

    node_type = bind_prop(
        "node_type",
        VariableInt,
        CameraNodeType.NORMAL,
        "Camera node type, see :class:`CameraNodeType` enum.",
    )
    test_widths = bind_prop_arr("test_widths", VariableInt)
    nodes = bind_prop_arr("c_node_ids", VariableUInt)
    control_widths = bind_prop_arr("control_widths", VariableVec2)
    zoom = bind_prop("zoom_h", VariableInt, 1080)
    width = bind_prop("width", VariableInt, 520)


class LevelEnd(Entity):
    """Level end flag class"""

    TYPE_IDENTIFIER = "level_end"

    def remap_ids(self, id_map: Dict[int, int]) -> None:
        """Remap entity IDs attached to this flag."""
        super().remap_ids(id_map)
        ents = self.entities
        for i, ent_id in enumerate(ents):
            ents[i] = id_map.get(ent_id, 0)

    entities = bind_prop_arr("ent_list", VariableUInt)
    finished = bind_prop("finished", VariableBool, False)


class ScoreBook(Entity):
    """Score book class"""

    TYPE_IDENTIFIER = "score_book"

    book_type = bind_prop("book_type", VariableString, b"")


class LevelDoor(Trigger):
    """Level door class"""

    TYPE_IDENTIFIER = "level_door"

    file_name = bind_prop("file_name", VariableString, b"")
    door_set = bind_prop("door_set", VariableUInt, 0)


class RedKeyDoor(Entity):
    """Red key door class"""

    TYPE_IDENTIFIER = "giga_gate"

    keys_needed = bind_prop("key_needed", VariableInt, 1)


class EntityHittable(Entity):
    """Base class for all 'hittable' types"""

    scale = bind_prop("dm_scale", VariableFloat, 1.0)

    def transform(self, mat: TxMatrix) -> None:
        """Adjust the entity scale"""
        super().transform(mat)
        self.scale = self.scale * mat.scale


class Enemy(EntityHittable):
    """Base class for all enemy types

    Subclasses can override :attr:`FILTH` to control how much "filth" is
    attributed to this entity. The DF file format requires a totalling of
    all filth in a level for completion calculations.
    """

    FILTH: int = 1


class EnemyLightPrism(Enemy):
    """Light prism entity class"""

    TYPE_IDENTIFIER = "enemy_tutorial_square"


class EnemyHeavyPrism(Enemy):
    """Heavy prism entity class. Note that although heavy prisms reward 3 dust
    when cleansing them they only count as 1 filth from the perspective of
    completion calculations."""

    TYPE_IDENTIFIER = "enemy_tutorial_hexagon"


class EnemySlimeBeast(Enemy):
    """Slime beast entity class"""

    TYPE_IDENTIFIER = "enemy_slime_beast"

    FILTH = 9


class EnemySlimeBarrel(Enemy):
    """Slime barrel (paint can) entity class"""

    TYPE_IDENTIFIER = "enemy_slime_barrel"

    FILTH = 3


class EnemySpringBall(Enemy):
    """Spring ball/blob entity class"""

    TYPE_IDENTIFIER = "enemy_spring_ball"

    FILTH = 5


class EnemySlimeBall(Enemy):
    """Slime ball (lab turkey) entity class"""

    TYPE_IDENTIFIER = "enemy_slime_ball"

    FILTH = 3


class EnemyTrashTire(Enemy):
    """Trash tire entity class"""

    TYPE_IDENTIFIER = "enemy_trash_tire"

    FILTH = 3

    max_fall_speed = bind_prop("max_fall_speed", VariableFloat, 800.0)


class EnemyTrashBeast(Enemy):
    """Trash beast (golem) entity class"""

    TYPE_IDENTIFIER = "enemy_trash_beast"

    FILTH = 9


class EnemyTrashCan(Enemy):
    """Trash can entity class"""

    TYPE_IDENTIFIER = "enemy_trash_can"

    FILTH = 9


class EnemyTrashBall(Enemy):
    """Trash ball entity class"""

    TYPE_IDENTIFIER = "enemy_trash_ball"

    FILTH = 3


class EnemyBear(Enemy):
    """Bear entity class"""

    TYPE_IDENTIFIER = "enemy_bear"

    FILTH = 9


class EnemyTotemLarge(Enemy):
    """Large totem (stoneboss) entity class"""

    TYPE_IDENTIFIER = "enemy_stoneboss"

    FILTH = 12


class EnemyTotemSmall(Enemy):
    """Totem entity class"""

    TYPE_IDENTIFIER = "enemy_stonebro"

    FILTH = 3


class EnemyPorcupine(Enemy):
    """Porcupine entity class"""

    TYPE_IDENTIFIER = "enemy_porcupine"


class EnemyWolf(Enemy):
    """Wolf entity class"""

    TYPE_IDENTIFIER = "enemy_wolf"

    FILTH = 5


class EnemyTurkey(Enemy):
    """Turkey (critter) entity class"""

    TYPE_IDENTIFIER = "enemy_critter"

    FILTH = 3


class EnemyFlag(Enemy):
    """Flag entity class"""

    TYPE_IDENTIFIER = "enemy_flag"

    FILTH = 5


class EnemyScroll(Enemy):
    """Scroll entity class"""

    TYPE_IDENTIFIER = "enemy_scrolls"


class EnemyTreasure(Enemy):
    """Treasure entity class"""

    TYPE_IDENTIFIER = "enemy_treasure"


class EnemyChestTreasure(Enemy):
    """Chest that spawns treasures entity class"""

    TYPE_IDENTIFIER = "enemy_chest_treasure"

    FILTH = 9


class EnemyChestScrolls(Enemy):
    """Chest that spawns scross entity class"""

    TYPE_IDENTIFIER = "enemy_chest_scrolls"

    FILTH = 9


class EnemyButler(Enemy):
    """Butler entity class"""

    TYPE_IDENTIFIER = "enemy_butler"


class EnemyMaid(Enemy):
    """Maid entity class"""

    TYPE_IDENTIFIER = "enemy_maid"


class EnemyKnight(Enemy):
    """Knign entity class"""

    TYPE_IDENTIFIER = "enemy_knight"

    FILTH = 9


class EnemyGargoyleBig(Enemy):
    """Big (punching) gargoyle entity class"""

    TYPE_IDENTIFIER = "enemy_gargoyle_big"

    FILTH = 5


class EnemyGargoyleSmall(Enemy):
    """Gargoyle (mansion turkey) entity class"""

    TYPE_IDENTIFIER = "enemy_gargoyle_small"

    FILTH = 3


class EnemyBook(Enemy):
    """Book entity class"""

    TYPE_IDENTIFIER = "enemy_book"

    FILTH = 3


class EnemyHawk(Enemy):
    """Hawk entity class"""

    TYPE_IDENTIFIER = "enemy_hawk"

    FILTH = 3


class EnemyKey(Enemy):
    """Key entity class"""

    TYPE_IDENTIFIER = "enemy_key"

    FILTH = 1

    def remap_ids(self, id_map: Dict[int, int]) -> None:
        """Remap the door ID."""
        super().remap_ids(id_map)
        self.door = id_map.get(self.door, 0)

    def transform(self, mat: TxMatrix) -> None:
        """Transform the last know coordinates."""
        super().transform(mat)

        self.lastX, self.lastY = mat.sample(self.lastX, self.lastY)

    door = bind_prop("door", VariableUInt, 0, "ID of door entity")
    lastX = bind_prop("lastKnowX", VariableFloat, 0.0)
    lastY = bind_prop("lastKnowY", VariableFloat, 0.0)


class EnemyDoor(Enemy):
    """Door entity class"""

    TYPE_IDENTIFIER = "enemy_door"

    FILTH = 0


class Apple(EntityHittable):
    """Apple entity class"""

    TYPE_IDENTIFIER = "hittable_apple"


class DustCharacter(EntityHittable):
    """Normal playable dust character entity types"""


class Dustman(DustCharacter):
    """Dustman entity class"""

    TYPE_IDENTIFIER = "dust_man"


class Dustgirl(DustCharacter):
    """Dustgirl entity classs"""

    TYPE_IDENTIFIER = "dust_girl"


class Dustkid(DustCharacter):
    """Dustkid entity class"""

    TYPE_IDENTIFIER = "dust_kid"


class Dustworth(DustCharacter):
    """Dustworth entity class"""

    TYPE_IDENTIFIER = "dust_worth"


class Dustwraith(DustCharacter):
    """Dustwraith entity class"""

    TYPE_IDENTIFIER = "dust_wraith"


class Leafsprite(DustCharacter):
    """Leaf sprite entity class"""

    TYPE_IDENTIFIER = "leaf_sprite"


class Trashking(DustCharacter):
    """Trash king entity class"""

    TYPE_IDENTIFIER = "trash_king"


class Slimeboss(DustCharacter):
    """Slime boss entity class"""

    TYPE_IDENTIFIER = "slime_boss"


_NOTES = """ Missing persist types and vars

custom_score_book
  level_list_id
dustmod_entity_tool
  dm_scale
  target_id
dustmod_trigger
  is_square
filth_switch
  height
  width
z_button
  entity_ids
  start_x
  start_y
z_char_force
  character
z_dash_trigger
  disable
z_kill_trigger
  kill
  only_dustman
  only_enemies
  stun
  stun_direction
  stun_force
  stun_radial
z_level_trigger
  filename
z_particle_trigger
  emitters
  particles
z_physics_trigger
  air_global
  attack_force_heavy
  attack_force_light
  d_dash_max
  d_skill_combo_max
  dash_speed
  di_global
  di_move_max
  di_speed
  di_speed_wall_lock
  fall_accel
  fall_max
  fric_global
  global
  heavy_fall_threshold
  hitrise_speed
  hop_a
  hover_accel
  hover_fall_threshold
  idle_fric
  jump_a
  land_fric
  roof_fric
  roof_run_length
  run_accel
  run_accel_over
  run_global
  run_max
  run_start
  skid_fric
  skid_threshold
  slope_max
  slope_slid_speed
  wall_run_length
  wall_slide_speed
z_respawner
  entity_ids
  max_respawns
  respawn_time
z_scale_trigger
  scale
z_string_list
  list
z_teleport_trigger
  tele_x
  tele_y
z_text_prop_trigger
  colour
  font
  font_size
  layer
  sublayer
  text
  text_rotation
  text_scale
z_wind_generator
  direction
  force
z_wind_trigger
  pressure_accelleration
  pressure_advection
  pressure_diffusion
  velocity_advection
  velocity_diffusion
  velocity_friction_a
  velocity_friction_b
  velocity_friction_c
  vorticity
"""



""" Module containing dustmaker's prop representation.  """
import math

from .transform import TxMatrix


class Prop:
    """Class respresenting a static prop in a map.

    To find appropriate values for :attr:`prop_set`, :attr:`prop_group`, and
    :attr:`prop_index` check out
    https://github.com/cmann1/PropUtils/tree/master/files/prop_reference.

    Attributes:
        layer_sub (int): The sublayer the prop is rendered on. Note that the prop
            layer is actually stored within the containing
            :class:`dustmaker.level.Level` itself.
        rotation (16-bit uint): Clockwise rotation of the prop ranging from 0 to 0xFFFF.
            0x4000 corresponds to a 90 degree rotation, 0x8000 to 180 degrees, 0xC000
            to 270 degrees. This rotation is logically applied after any flips have
            been applied.
        flip_x (bool): Flip the prop horizontally
        flip_y (bool): Flip the prop vertically
        scale (float): Prop scaling factor. This is only available in `LevelType.DUSTMOD`
            type maps and is fairly coarse in the resolution of the scaling factor.
        prop_set (int): Identifier indicating what prop set this prop comes from. This
            appears to match :class:`dustmaker.tile.TileSpriteSet`.
        prop_group (int): Identifier indicating what prop group this prop comes from.
        prop_index (int): Index of the desired prop sprite.
        palette (int): The colour variant of the prop to render.
    """

    def __init__(
        self,
        layer_sub: int,
        rotation: int,
        flip_x: bool,
        flip_y: bool,
        scale: float,
        prop_set: int,
        prop_group: int,
        prop_index: int,
        palette: int,
    ) -> None:
        self.layer_sub = layer_sub
        self.rotation = rotation
        self.flip_x = flip_x
        self.flip_y = flip_y
        self.scale = scale
        self.prop_set = prop_set
        self.prop_group = prop_group
        self.prop_index = prop_index
        self.palette = palette

    def transform(self, mat: TxMatrix) -> None:
        """
        Performs the requested transformation on the prop's :attr:`rotation` and
        :attr:`flip_y` attributes.
        """
        self.rotation = self.rotation - int(0x10000 * mat.angle / math.pi / 2) & 0xFFFF
        if mat.flipped:
            self.flip_y = not self.flip_y
            self.rotation = -self.rotation & 0xFFFF



""" Module defining TxMatrix class used for affine transformations. """
import math
from typing import Iterable, Tuple, Union


class TxMatrix(tuple):
    """Immutable transformation matrix. This is a subclass of tuple that
    enforce itself to be a 3x3 affine transformation matrix.
    """

    IDENTITY: "TxMatrix"
    HFLIP: "TxMatrix"
    VFLIP: "TxMatrix"
    ROTATE: Tuple["TxMatrix", "TxMatrix", "TxMatrix", "TxMatrix"]

    def __new__(cls, matrix: Iterable[Iterable[float]]) -> "TxMatrix":
        mat = super().__new__(cls, (tuple(row) for row in matrix))  # type: ignore
        assert len(mat) == 3, "expected 3x3 matrix"
        assert all(len(row) == 3 for row in mat), "expected 3x3 matrix"
        assert all(
            isinstance(elem, (int, float)) for row in mat for elem in row
        ), "all elements must be ints or floats"
        assert mat[2] == (0, 0, 1), "last row must always be [0, 0, 1]"
        return mat

    def translate(self, x: float, y: float) -> "TxMatrix":
        """Add an additional translation to the transformation matrix.
        mat.translate(x, y) == TxMatrix.IDENTITY.translate(x, y) * mat
        """
        return TxMatrix(
            (
                (self[0][0], self[0][1], self[0][2] + x),
                (self[1][0], self[1][1], self[1][2] + y),
                (0, 0, 1),
            )
        )

    @property
    def angle(self) -> float:
        """Returns the amount a vertical line has been rotated clockwise
        in radians under this transformation.
        """
        return math.atan2(self[1][1], self[1][0]) - math.pi / 2

    @property
    def determinant(self) -> float:
        """Get the determinant of the transformation matrix"""
        return self[0][0] * self[1][1] - self[0][1] * self[1][0]

    @property
    def flipped(self) -> bool:
        """Returns True if the transformation matrix flips an axis. This is the
        same as the :attr:`determinant` being negative."""
        return self.determinant < 0

    @property
    def scale(self) -> float:
        """Returns the scaling of the transformation matrix. This is just the
        square root of the absolute value of the determinant which itself
        represents how much 2D areas expand in scale."""
        return math.sqrt(abs(self.determinant))

    def __mul__(self, rhs: Union[float, "TxMatrix"]) -> "TxMatrix":
        if isinstance(rhs, (int, float)):
            # Multiply by a scalar
            return TxMatrix(
                (
                    (self[0][0] * rhs, self[0][1] * rhs, self[0][2] * rhs),
                    (self[1][0] * rhs, self[1][1] * rhs, self[1][2] * rhs),
                    (0, 0, 1),
                )
            )

        # Matrix multiplication optimized for 2D affine matrixes.
        ml = self
        mr = rhs
        return TxMatrix(
            (
                (
                    ml[0][0] * mr[0][0] + ml[0][1] * mr[1][0],
                    ml[0][0] * mr[0][1] + ml[0][1] * mr[1][1],
                    ml[0][0] * mr[0][2] + ml[0][1] * mr[1][2] + ml[0][2],
                ),
                (
                    ml[1][0] * mr[0][0] + ml[1][1] * mr[1][0],
                    ml[1][0] * mr[0][1] + ml[1][1] * mr[1][1],
                    ml[1][0] * mr[0][2] + ml[1][1] * mr[1][2] + ml[0][2],
                ),
                (0, 0, 1),
            )
        )

    def __rmul__(self, lhs: Union[float]) -> "TxMatrix":
        return self * lhs

    def sample(self, x: float, y: float, *, with_offset=True) -> Tuple[float, float]:
        """Sample the transformed coordinates of (x, y)"""
        return (
            self[0][0] * x + self[0][1] * y + (self[0][2] if with_offset else 0),
            self[1][0] * x + self[1][1] * y + (self[1][2] if with_offset else 0),
        )


#: The identity matrix
TxMatrix.IDENTITY = TxMatrix(((1, 0, 0), (0, 1, 0), (0, 0, 1)))

#: Matrix representing a horizontal flip
TxMatrix.HFLIP = TxMatrix(((-1, 0, 0), (0, 1, 0), (0, 0, 1)))

#: Matrix representing a vertical flip
TxMatrix.VFLIP = TxMatrix(((1, 0, 0), (0, -1, 0), (0, 0, 1)))

_CS = (1, 0, -1, 0)

#: ROTATE[i] gives the rotation matrix for rotating by 90-degrees
#: clockwise `i` times.
TxMatrix.ROTATE = tuple(  # type: ignore
    TxMatrix(((_CS[i], _CS[3 - i], 0), (_CS[(3 + i) & 0x3], _CS[i], 0), (0, 0, 1)))
    for i in range(4)
)



"""
Module providing methods for write Dustforce binary formats including level
files.
"""
import io
from math import floor, log
from typing import Dict, Iterable, List, Tuple, Union
import zlib

from .bitio import BitIOWriter
from .level import Level
from .exceptions import LevelParseException
from .tile import TileSpriteSet
from .variable import (
    Variable,
    VariableArray,
    VariableBool,
    VariableFloat,
    VariableInt,
    VariableString,
    VariableStruct,
    VariableType,
    VariableUInt,
    VariableVec2,
)


class _LevelSegment:
    """Container class with raw information about a segment"""

    def __init__(self):
        self.tiles = [[] for _ in range(256)]
        self.entities = []
        self.props = []
        self.present = True


class _LevelRegion:
    """Container class for region data"""

    def __init__(self):
        self.segment_map = {}
        self.backdrop = _LevelSegment()
        self.backdrop.present = False

    def get_segment(self, x: int, y: int) -> _LevelSegment:
        """Return the segment at the given coordinate within this region"""
        seg_key = ((x // 16) & 0xF, (y // 16) & 0xF)

        seg = self.segment_map.get(seg_key)
        if seg is not None:
            return seg

        seg = _LevelSegment()
        self.segment_map[seg_key] = seg
        return seg


class _RegionMap:
    """Container class for region/segment information."""

    def __init__(self):
        self.region_map = {}

    def get_region(self, x: int, y: int) -> _LevelRegion:
        """Return the region at the given coordinate"""
        reg_ky = (x // 256, y // 256)
        reg = self.region_map.get(reg_ky)
        if reg is not None:
            return reg

        reg = _LevelRegion()
        self.region_map[reg_ky] = reg
        return reg

    def get_segment(self, x: int, y: int) -> _LevelSegment:
        """Return the segment at the given coordinate"""
        return self.get_region(x, y).get_segment(x, y)

    def get_region_keys(self) -> List[Tuple[int, int]]:
        """Return a list of the region keys sorted in the order required by the
        Dustforce level format.
        """

        def norm_for_sort(coord):
            """Do some stuff to order regions correctly"""
            x = coord[0]
            y = coord[1]
            if x < 0:
                x += 1 << 16
            if y < 0:
                y += 1 << 16
            return (x, y)

        return sorted(self.region_map, key=norm_for_sort)


def _compute_region_map(level: Level) -> _RegionMap:
    """Constructs a region map based on the passed level file. This
    structure separates all the tiles/props/entities into their respective
    regions as is used in the underlying binary representation.
    """
    rmap = _RegionMap()
    for (layer, tx, ty), tile in level.tiles.items():
        rmap.get_segment(tx, ty).tiles[layer].append((tx & 0xF, ty & 0xF, tile))

    for id_num, (x, y, entity) in level.entities.items():
        rmap.get_segment(int(x / 48), int(y / 48)).entities.append(
            (id_num, x, y, entity)
        )

    for id_num, (layer, x, y, prop) in level.props.items():
        rmap.get_segment(int(x / 48), int(y / 48)).props.append(
            (id_num, layer, x, y, prop)
        )

    if level.backdrop is None:
        return rmap

    for (layer, x, y), tile in level.backdrop.tiles.items():
        seg = rmap.get_region(x * 16, y * 16).backdrop
        seg.present = True
        seg.tiles[layer].append((x & 0xF, y & 0xF, tile))

    for id_num, (layer, x, y, prop) in level.backdrop.props.items():
        seg = rmap.get_region(int(x / 48), int(y / 48)).backdrop
        seg.present = True
        seg.props.append((id_num, layer, x, y, prop))

    return rmap


class DFWriter(BitIOWriter):
    """Helper class to write Dustforce binary files"""

    def write_float(self, ibits: int, fbits: int, val: float) -> None:
        """Write a float `val` to the output stream

        Arguments:
            ibits (int): Number of integer bits
            fbits (int): Number of fractional bits
            val (float): The floating point number to write
        """
        ipart = abs(floor(val))
        fpart = int((val - floor(val)) * (2 ** (fbits - 1)))
        self.write(1, 1 if val < 0 else 0)
        self.write(ibits - 1, ipart)
        self.write(fbits, fpart)

    def write_6bit_str(self, text: str) -> None:
        """Write a '6-bit' string. These are strings with length between
        0 and 63 inclusive that contain only alpha-numeric lower and uppercase
        characters in addition to '_' and '{'.

        Raises:
            ValueError: If length of string exceeds 63
            ValueError: If invalid character is present
        """
        if len(text) > 63:
            raise ValueError("6-bit str too long")

        self.write(6, len(text))
        for ch in text:
            v = ord(ch)
            if ord("0") <= v <= ord("9"):
                self.write(6, v - ord("0"))
            elif ord("A") <= v <= ord("Z"):
                self.write(6, v - ord("A") + 10)
            elif ord("a") <= v <= ord("z"):
                self.write(6, v - ord("a") + 37)
            elif ch == "_":
                self.write(6, 36)
            elif ch == "{":
                self.write(6, 63)
            else:
                raise ValueError("invalid character in 6-bit string")

    def write_variable(self, var: Variable) -> None:
        """Write a variable to the output stream. This does not write the
        type of the variable, that will need to be encoded somewhere else if
        needed.

        Arguments:
            var (Variable): The variable to write to the stream.

        Raises:
            ValueError: If a VariableString value is longer than 65535 bytes.
                Note that structs and arrays each handle their string sub-elements
                in a way that they may be longer than this limit.
            ValueError: If a VariableArray has more than 65535 elements including
                any continuations if the element type is VariableString.
            LevelParseException: If `var` is of unknown variable type.
        """
        max_width = (1 << 16) - 1

        value = var.value
        if isinstance(var, VariableBool):
            self.write(1, 1 if value else 0)
            return
        if isinstance(var, (VariableUInt, VariableInt)):
            self.write(32, value)
            return
        if isinstance(var, VariableFloat):
            self.write_float(32, 32, value)
            return

        if isinstance(var, VariableString):
            if len(var.value) > max_width:
                raise ValueError("VariableString length too long")

            self.write(16, len(var.value))
            self.write_bytes(var.value)
            return

        if isinstance(var, VariableVec2):
            self.write_float(32, 32, value[0])
            self.write_float(32, 32, value[1])
            return

        if isinstance(var, VariableArray):
            atype, arr = value
            arrlen = len(arr)
            if atype is VariableString:
                for x in arr:
                    arrlen += len(x.value) // max_width
            if arrlen > max_width:
                raise ValueError("VariableArray length too long")

            self.write(4, atype._vtype)
            self.write(16, arrlen)
            if atype is not VariableString:
                for x in arr:
                    self.write_variable(x)
            else:
                for x in arr:
                    # Write array continuations
                    xs = x.value
                    for i in range(0, 1 + len(xs) // max_width):
                        self.write_variable(
                            VariableString(xs[i * max_width : (i + 1) * max_width])
                        )
            return

        if isinstance(var, VariableStruct):
            for elem_key, elem_var in var.value.items():
                if not isinstance(elem_var, VariableString):
                    self.write(4, elem_var._vtype)
                    self.write_6bit_str(elem_key)
                    self.write_variable(elem_var)
                    continue

                # Write struct string continuations
                xs = elem_var.value
                for i in range(0, 1 + len(xs) // max_width):
                    self.write(4, elem_var._vtype)
                    self.write_6bit_str(elem_key)
                    self.write_variable(
                        VariableString(xs[i * max_width : (i + 1) * max_width])
                    )

            self.write(4, VariableType.NULL)

            return

        raise LevelParseException("unknown var type")

    def _write_segment(self, seg_x: int, seg_y: int, segment: _LevelSegment) -> None:
        """Write a segment to the output stream"""
        assert self.aligned()

        start_index = self.bit_tell()
        self.bit_seek(start_index + 200)

        flags = 0
        tiles = [(i, segment.tiles[i]) for i in range(21) if segment.tiles[i]]

        (dust_filth, enemy_filth, tile_surface, dustblock_filth) = (0, 0, 0, 0)

        dusts = []
        if tiles:
            flags |= 1

            self.write(8, len(tiles))
            for layer, tilelayer in tiles:
                self.write(8, layer)
                self.write(10, len(tilelayer))

                for x, y, tile in sorted(tilelayer, key=lambda x: (x[1], x[0])):
                    if layer == 19 and tile.has_filth():
                        dusts.append((x, y, tile))
                    if layer == 19:
                        if tile.is_dustblock():
                            dustblock_filth += 1
                        for edge in tile.edge_data:
                            if edge.filth_spike:
                                continue

                            if edge.filth_sprite_set != TileSpriteSet.NONE_0:
                                dust_filth += 1

                            if edge.solid and edge.visible:
                                tile_surface += 1

                    self.write(5, x)
                    self.write(5, y)
                    self.write(5, tile.shape)
                    self.write(3, tile.tile_flags)
                    self.write_bytes(tile._pack_tile_data())

        if dusts:
            flags |= 2

            self.write(10, len(dusts))
            for (x, y, tile) in dusts:
                self.write(5, x)
                self.write(5, y)
                self.write_bytes(tile._pack_dust_data())

        if segment.props:
            flags |= 8

            self.write(16, len(segment.props))
            for id_num, layer, x, y, prop in segment.props:
                self.write(32, id_num)
                self.write(8, layer)
                self.write(8, prop.layer_sub)

                scale_lg = int(round(log(prop.scale) / log(50.0) * 24.0)) + 32
                x_scale = (scale_lg // 7) ^ 0x4
                y_scale = (scale_lg % 7) ^ 0x4
                x_int = int(abs(x))
                y_int = int(abs(y))
                x_sgn = 1 if x < 0 else 0
                y_sgn = 1 if y < 0 else 0
                self.write(1, x_sgn)
                self.write(27, x_int)
                self.write(4, x_scale)
                self.write(1, y_sgn)
                self.write(27, y_int)
                self.write(4, y_scale)

                self.write(16, prop.rotation)
                self.write(1, 1 if prop.flip_x else 0)
                self.write(1, 1 if prop.flip_y else 0)
                self.write(8, prop.prop_set)
                self.write(12, prop.prop_group)
                self.write(12, prop.prop_index)
                self.write(8, prop.palette)

        if segment.entities:
            flags |= 4

            extra_names_io = io.BytesIO()
            extra_names_writer = DFWriter(extra_names_io)
            self.write(16, len(segment.entities))
            for (id_num, x, y, entity) in segment.entities:
                enemy_filth += getattr(entity, "FILTH", 0)
                self.write(32, id_num)
                if entity.etype.startswith("z_") or entity.etype == "entity":
                    self.write_6bit_str("entity")
                    extra_names_writer.write_6bit_str(entity.etype)
                else:
                    self.write_6bit_str(entity.etype)
                self.write_float(32, 8, x)
                self.write_float(32, 8, y)
                self.write(16, entity.rotation)
                self.write(8, entity.layer)
                self.write(1, 0 if entity.flip_x else 1)
                self.write(1, 0 if entity.flip_y else 1)
                self.write(1, 1 if entity.visible else 0)
                self.write_variable(VariableStruct(entity.variables))

            extra_names_writer.align()
            entity_names_pos = self.bit_tell()
            self.write_bytes(extra_names_io.getvalue())
            self.align()
            self.write(32, self.bit_tell() - entity_names_pos)

        self.align()

        end_index = self.bit_tell()
        self.bit_seek(start_index)

        self.write(32, (end_index - start_index) // 8)
        self.write(16, 8)
        self.write(8, seg_x)
        self.write(8, seg_y)
        self.write(8, 16)

        self.write(32, 0)
        self.write(16, dust_filth)
        self.write(16, enemy_filth)

        self.write(16, tile_surface)
        self.write(16, dustblock_filth)

        self.write(32, flags)
        self.bit_seek(end_index)

    def _write_region(self, x: int, y: int, region: _LevelRegion) -> None:
        """Writes a level region to the output stream"""
        segments_io = io.BytesIO()
        with DFWriter(segments_io) as segments_writer:
            for coord, segment in sorted(region.segment_map.items()):
                segments_writer._write_segment(coord[0], coord[1], segment)
            if region.backdrop.present:
                segments_writer._write_segment(0, 0, region.backdrop)
            uncompressed_data = segments_io.getvalue()

        compressed_data = zlib.compress(uncompressed_data)

        self.write(32, 17 + len(compressed_data))
        self.write(32, len(uncompressed_data))
        self.write(16, x)
        self.write(16, y)
        self.write(16, 14)
        self.write(16, len(region.segment_map))
        self.write(8, 1 if region.backdrop.present else 0)
        self.write_bytes(compressed_data)

    # Should match the number of bits written by write_metadata
    _METADATA_BIT_SIZE = 224

    def _write_metadata(self, level: Level) -> None:
        """Write a level metadata block to the output stream."""
        self.write_bytes(b"DF_MTD")
        self.write(16, 4)
        self.write(32, 0)
        self.write(32, level.calculate_max_id(reset=False) + 1)
        self.write(32, 0)
        self.write(32, 0)
        self.write(32, 0)

    def write_var_file(self, header: bytes, var_data: Dict[str, Variable]) -> None:
        """Writes a variable mapping with a given header. There are several
        file types that Dustforce use that are expressed this way including
        notably "stats1" (header=b"DF_STA") and "config" (header=b"DF_CFG").

        Arguments:
            header (bytes): The file header at the start of the stream.
        """
        assert self.aligned()

        start_index = self.bit_tell()
        self.bit_seek(start_index + 8 * len(header) + 48)

        self.write_variable(VariableStruct(var_data))
        self.align()
        end_index = self.bit_tell()

        self.bit_seek(start_index)

        self.write_bytes(header)
        self.write(16, 1)
        self.write(32, (end_index - start_index) // 8)
        self.bit_seek(end_index)

    def write_level_ex(
        self,
        level: Level,
        region_offsets: List[int],
        region_data: Union[bytes, Iterable[bytes]],
    ) -> None:
        """Extended version of :meth:`write_level`.

        Writes `level` to the output stream. This is the advanced API for
        efficiently modifiying level metadata. If `region_offsets` is non-empty
        this will use `region_offsets` and `region_data` to populate the region
        data section of the output rather than calculating it from `level`.

        See :meth:`dustmaker.dfreader.DFReader.read_level_ex` for examples
        on how to use this in practice.

        Arguments:
            level (Level): The level file to write
            region_offsets (list[int]): Region offset metadata as returned by
                :meth:`DFReader.read_level_ex`. If empty this behaves the same
                as :meth:`write_level`. If you wish to omit all region data pass
                `[0]` instead.
            region_data: (bytes | Iterable[bytes]): Bytes data (or an iterable
                of bytes data) to write in the region section of the map. If
                `region_offsets` is empty this argument is ignored.
        """
        assert self.aligned()

        start_index = self.bit_tell()
        header_size = 160 + self._METADATA_BIT_SIZE + len(level.sshot) * 8

        self.bit_seek(start_index + header_size)

        # Write variable mapping and record size
        self.write_variable(VariableStruct(level.variables))

        # Skip over the region directory for now
        if region_offsets:
            num_regions, end_index = self._write_regions_raw(
                region_offsets, region_data
            )
        else:
            num_regions, end_index = self._write_regions_from_level(level)

        # Got back to start and rewrite header.
        self.bit_seek(start_index)
        self.write_bytes(b"DF_LVL")
        self.write(16, 44)
        self.write(32, (end_index - start_index) // 8)
        self.write(32, num_regions)
        self._write_metadata(level)
        self.write(32, len(level.sshot))
        self.write_bytes(level.sshot)
        self.bit_seek(end_index)

    def _write_regions_raw(
        self, region_offsets: List[int], region_data: Union[bytes, Iterable[bytes]]
    ) -> Tuple[int, int]:
        """Helper function to write region data from raw offset/data
        information.

        Returns:
            The ending bit index.
        """
        for region_offset in region_offsets[:-1]:
            self.write(32, region_offset)

        self.align()

        start_index = self.bit_tell()
        if isinstance(region_data, bytes):
            self.write_bytes(region_data)
        else:
            for data in region_data:
                self.write_bytes(data)
        end_index = self.bit_tell()

        if end_index - start_index != region_offsets[-1] * 8:
            raise LevelParseException("got unexpected amount of regiond ata")

        return len(region_offsets) - 1, self.bit_tell()

    def _write_regions_from_level(self, level: Level) -> Tuple[int, int]:
        """Helper function to write region data from the level

        Returns:
            The ending bit index.
        """
        rmap = _compute_region_map(level)
        region_dir_index = self.bit_tell()
        region_data_index = (region_dir_index + 32 * len(rmap.region_map) + 7) & ~0x7
        self.bit_seek(region_data_index)
        region_offsets = []

        # Write the region data
        for rx, ry in rmap.get_region_keys():
            self.align()
            region_offsets.append((self.bit_tell() - region_data_index) // 8)
            self._write_region(rx, ry, rmap.region_map[(rx, ry)])

        end_index = self.bit_tell()

        # Go back and write the region directory. This might not be
        # byte aligned which doesn't can zero previous bits in the partial byte.
        # Since the first element is zero anyway we just skip it to avoid
        # writing over the var map.
        self.bit_seek(region_dir_index + 32, allow_unaligned=True)
        for region_offset in region_offsets[1:]:
            self.write(32, region_offset)

        return len(rmap.region_map), end_index

    def write_level(self, level: Level) -> None:
        """Writes `level` to the output stream. This is equivalent to
        `write_level_ex(level, [], b"")`.

        Arguments:
            level (Level): The level file to write
        """
        self.write_level_ex(level, [], b"")


def write_level(level: Level) -> bytes:
    """Convenience function to write a map file directly to bytes in memory.

    Arguments:
        level (Level): The level file to write

    Returns:
        The bytes that encode that level file
    """
    with DFWriter(io.BytesIO()) as writer:
        writer.write_level(level)
        return writer.data.getvalue()



"""
Module defining the primary interface for working with levels in dustmaker.
"""
import copy
from enum import IntEnum
import functools
import math
from typing import Callable, Dict, Optional, Tuple, TypeVar

from .entity import bind_prop, Entity
from .exceptions import LevelException
from .prop import Prop
from .tile import (
    Tile,
    TileEdgeData,
    TileShape,
    TileSide,
    TileSpriteSet,
    SHAPE_VERTEXES,
    SIDE_CLOCKWISE_INDEX,
)
from .transform import TxMatrix
from .variable import Variable, VariableBool, VariableInt, VariableString

T = TypeVar("T")


class LevelType(IntEnum):
    """Enum defining the different level types."""

    NORMAL = 0
    NEXUS = 1
    NEXUS_MP = 2
    KOTH = 3
    SURVIVAL = 4
    DUSTMOD = 6


class _LateBoundDescriptor:
    """Utility class to late bind a property descriptor to an instance"""

    def __init__(self, attrname: str, doc: str) -> None:
        self.attrname = attrname
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return getattr(obj, self.attrname).__get__(obj, objtype)

    def __set__(self, obj, value):
        return getattr(obj, self.attrname).__set__(obj, value)

    def __delete__(self, obj):
        return getattr(obj, self.attrname).__delete__(obj)


class PlayerPosition:
    """Used internally to manage player position accessors. Meant to be used
    through accesss to :meth:`Level.start_position`."""

    def __init__(self, variables: Dict[str, Variable], player: int):
        self.variables = variables
        self._x = bind_prop(f"p{player}_x", VariableInt, 0)
        self._y = bind_prop(f"p{player}_y", VariableInt, 0)

    x = _LateBoundDescriptor("_x", "int: Player start x-coordinate in pixels")
    y = _LateBoundDescriptor("_y", "int: Player start y-coordinate in pixels")


class Level:
    """Represents a Dustforce level/map or the backdrop to its `parent` level. If
    this is a backdrop then :attr:`parent` will be set to the parent
    :class:`Level` object.

    Attributes:
        parent (Level, optional): For backdrops this is the containing level.
            Otherwise this is set to None.
        backdrop (Level, optional): The backdrop Level object or None if this is a
            backdrop level. Backdrop levels are scaled up 16x from the parent
            level's coordinate system and should only contain tiles and props.
        tiles (dict[(int, int, int), Tile]): A dict mapping (layer, x, y)
            to Tile objects.
        props (dict[int, (int, float, float, Prop)]): A dict mapping prop ids to
            (layer, x, y, Prop) tuples.
        entities (dict[int, (float, float, Entity)]): A dict mapping entity
            ids to (x, y, Entity) tuples. Ignored for backdrops.
        variables (dict[str, Variable]): A raw mapping of string keys to Variable objects.
            Some of these variables have nicer accessor properties like :attr:`name` but may
            be accessed raw through this dictionary. Ignored for backdrops.
        sshot (bytes): The level thumbnail image as a PNG binary. Ignored for
            backdrops.
    """

    def __init__(self, *, parent: Optional["Level"] = None) -> None:
        self._next_id = 100
        self.tiles: Dict[Tuple[int, int, int], Tile] = {}
        self.props: Dict[int, Tuple[int, float, float, Prop]] = {}
        self.variables: Dict[str, Variable] = {}

        self.parent = parent
        self.sshot = b""
        self.entities: Dict[int, Tuple[float, float, Entity]] = {}

        self.backdrop: Optional["Level"] = (
            Level(parent=self) if parent is None else None
        )
        self.dustmod_version = b"dustmaker"

    def _gen_id(self) -> int:
        """Allocate and return an ID for a new entity or prop."""
        if self.parent is not None:
            return self.parent._gen_id()
        result = self._next_id
        self._next_id += 1
        return result

    def _note_id(self, id_num: int) -> None:
        """Called to update the internally tracked minimum ID."""
        if self.parent is not None:
            self.parent._note_id(id_num)
        else:
            self._next_id = max(self._next_id, id_num + 1)

    name = bind_prop("level_name", VariableString, b"")
    virtual_character = bind_prop("vector_character", VariableBool, False)
    level_type = bind_prop(
        "level_type",
        VariableInt,
        LevelType.NORMAL,
        "Level type of the level, see :class:`LevelType` enum.",
    )
    dustmod_version = bind_prop("dustmod_version", VariableString, b"")

    def start_position(self, player: int = 1) -> PlayerPosition:
        """Access and modify the starting position of each player.

        Args:
            player (int, optional): The player to access the starting position
                of. Valid options are 1, 2, 3, 4. Defaults to player 1.

        Returns:
            An accessor class with `x` and `y` attributes that can be get/set.
        """
        return PlayerPosition(self.variables, player)

    def add_prop(
        self, layer: int, x: float, y: float, prop: Prop, id_num: Optional[int] = None
    ) -> int:
        """Adds a new :class:`Prop` to the level and returns its id. This is the
        preferred way of adding props to a level. Do not directly add props to the
        :attr:`props` attribute and use this method as it may overwrite props.

        Args:
            x (float): The x position of the prop.
            y (float): The y position of the prop.
            prop (Prop): The Prop object to add to the level.
            id_num: (int, optional): The prop identifier. If not set the identifier
                will be allocated for you.

        Returns:
            The ID of the newly added prop.

        Raises:
            LevelException: If the given ID is already in use.
        """
        if id_num is None:
            id_num = self._gen_id()
        else:
            self._note_id(id_num)
        if id_num in self.props:
            raise LevelException("level already has prop id")
        self.props[id_num] = (layer, x, y, prop)
        return id_num

    def add_entity(
        self, x: float, y: float, entity: Entity, id_num: Optional[int] = None
    ) -> int:
        """Adds a new :class:`Entity` to the level and returns its id. This is the
        preferred way of adding entities to a level. Do not directly add entities to the
        :attr:`entities` attribute and use this method as it may overwrite entities.

        Args:
            x (float): The x position of the entity.
            y (float): The y position of the entity.
            entity (Entity): The Entity object to add to the level.
            id_num: (int, optional): The entity identifier. If not set the
                identifier will be allocated for you.

        Returns:
            The ID of the newly added entity.

        Raises:
            LevelException: If the given ID is already in use.
        """
        if id_num is None:
            id_num = self._gen_id()
        else:
            self._note_id(id_num)

        if id_num in self.entities:
            raise LevelException("level already has entity with ID")
        self.entities[id_num] = (x, y, entity)
        return id_num

    def translate(self, x: float, y: float) -> None:
        """Translate (move) the entire level.  This is just a convenience method
        around :meth:`transform`.

        Args:
            x (float): The number of pixels to move horizontally.
            y (float): The number of pixels to move vertically.
        """
        self.transform(TxMatrix.IDENTITY.translate(x, y))

    def remap_ids(self, min_id: int = 100) -> None:
        """Remap prop and entity ids starting at `min_id`. This is useful
        when attempting to merge two levels to keep their ID space separate.
        Do not call directly on a backdrop level.

        Args:
            min_id (int, optional): The minimum ID to assign to a prop or entity.

        Warning:
            Dustmaker has no way to automatically remap entity IDs in script persist data.
        """
        self._next_id = min_id

        prop_remap = {id_num: self._gen_id() for id_num in self.props}
        self.props = {prop_remap[id_num]: prop for id_num, prop in self.props.items()}

        entity_remap = {id_num: self._gen_id() for id_num in self.entities}
        self.entities = {
            entity_remap[id_num]: entity for id_num, entity in self.entities.items()
        }
        for _, _, entity in self.entities.values():
            entity.remap_ids(entity_remap)

        if self.backdrop is not None:
            self.backdrop.remap_ids()

    def calculate_max_id(self, reset: bool = True) -> int:
        """Calculates the maximum prop or entity ID currently in use. This will
        always return at least 100 due to Dustforce reserving many of the lower
        IDs for special entities like the camera.

        Arguments:
            reset (bool, optional): If set (the default) the internal next ID
                allocator will be reset based off this result. Otherwise the max
                ID will be at least one less than the next ID.
        """
        init = 100 if reset else self._next_id - 1
        mx_id = max(max(self.props, default=init), max(self.entities, default=init))
        if self.backdrop is not None:
            mx_id = max(mx_id, self.backdrop.calculate_max_id())
        self._next_id = mx_id + 1
        return mx_id

    def merge(self, other_level: "Level", remap_ids: bool = True) -> None:
        """Merge another level into this one.

        Args:
            other_map (Level): The level to merge into this one.
            remap_ids (bool, optional): Wether to remap the ID space of each
                level so they do not interfere with each other. This is True by
                default.
        """
        if remap_ids:
            self.remap_ids(other_level.calculate_max_id() + 1)

        self.tiles.update(copy.deepcopy(other_level.tiles))
        self.props.update(copy.deepcopy(other_level.props))
        self.entities.update(copy.deepcopy(other_level.entities))
        if self.backdrop is not None and other_level.backdrop is not None:
            self.backdrop.merge(other_level.backdrop, remap_ids=False)

    def transform(self, mat: TxMatrix) -> None:
        """Transforms the level with the given affine transformation matrix.  Note
        that this will probably not produce desirable results if the
        transformation matrix is not some mixture of a translation, flip,
        and 90 degree rotations. Use :meth:`upscale` if you wish to performan
        an upscale as well as a transformation.

        In most cases you can use one of :meth:`flip_horizontal`,
        :meth:`flip_vertical:, :meth:`rotate`, :meth:`translate`, :meth:`upscale`
        instead of this method.

        Args:
            mat: The 3 by 3 affine transformation matrix [x', y', 1]' = mat *
                [x, y, 1]'. It should be of the form
                `mat = [[xx, xy, ox], [yx, yy, oy], [0, 0, 1]]`.

        Warning:
            Dustmaker has no way to automatically transform positional data in
            script persist data.
        """

        def _transform_tiles():
            for (layer, x, y), tile in self.tiles.items():
                tile.transform(mat)
                tx, ty = mat.sample(x * 48, y * 48)

                yield (layer, int(round(tx / 48.0)), int(round(ty / 48.0))), tile

        def _transform_props():
            for id_num, (layer, x, y, prop) in self.props.items():
                prop.transform(mat)
                tx, ty = mat.sample(x, y)
                yield id_num, (layer, tx, ty, prop)

        def _transform_entities():
            for id_num, (x, y, entity) in self.entities.items():
                entity.transform(mat)
                tx, ty = mat.sample(x, y)
                yield id_num, (tx, ty, entity)

        self.tiles = dict(_transform_tiles())
        self.props = dict(_transform_props())
        self.entities = dict(_transform_entities())

        for player in range(1, 5):
            pos = self.start_position(player)
            pos_tx, pos_ty = mat.sample(pos.x, pos.y)
            pos.x, pos.y = int(round(pos_tx)), int(round(pos_ty))

        if self.backdrop is not None:
            self.backdrop.transform(mat)

    def flip_horizontal(self) -> None:
        """Flips the level horizontally. This is a convenience function around
        :meth:`transform`."""
        self.transform(TxMatrix.HFLIP)

    def flip_vertical(self) -> None:
        """Flips the level vertically. This is a convenience function around
        :meth:`transform`."""
        self.transform(TxMatrix.VFLIP)

    def rotate(self, times: int = 1) -> None:
        """Rotates the level 90 degrees clockwise. This is a convenience function
        around :meth:`transform`.

        Args:
            times (int, optional): The number of 90 degree clockwise rotations
                to perform. `times` may be negative to perform counterclockwise
                rotations.
        """
        self.transform(TxMatrix.ROTATE[times % 4])

    def upscale(self, factor: int, *, mat: TxMatrix = TxMatrix.IDENTITY) -> None:
        """Increase the size of the level along each axis.

        Args:
            factor (int): The scaling factor (>1). e.g. if `factor` = 2 each tile
                tile will be represented by a 2x2 tile squre in the resulting
                level.
            mat (optional): An additional transformation matrix to pass
                to :meth:`transform` along with the upscaling.
        """
        self.transform(mat * factor)
        self.tiles = {
            (layer, x + dx, y + dy): ntile
            for (layer, x, y), tile in self.tiles.items()
            for dx, dy, ntile in tile.upscale(factor)
        }

    def calculate_edge_visibility(
        self,
        *,
        visible_callback: Optional[
            Callable[[int, int, TileSide, Tile, Tile], bool]
        ] = None,
    ) -> None:
        """This method will automatically calculate edge solidity and
        visibility in a way meant to match Dustforce rules.

        Solidity will always imply visibility. An edge that doesn't exist
        for the given tile shape is always not visible. Otherwise an edge
        that is not flush to a
        tile border (e.g. the diagonal of a slope tile) is solid. Otherwise
        if the neighboring side does not exist or is not also flush the
        edge is solid.

        In any other case the edge is not solid. If the tile sprite information
        matches its neighbor the edge is not visible. Otherwise
        `visible_callback` is called to determine if the edge is visible. If
        `visible_callback` is not set this defaults to the edge being visible
        if it's a bototm or right edge.
        bottom or right edge.

        Arguments:
            visible_callback (Callable): Callback used to determine if a given
                edge should be visible if all other checks have passed. Typically
                this function should be anti-symetric so that there are not
                overlapping visible edges. Called as
                `visible_callback(x, y, side, tile, neighbor_tile)`.
        """
        neighbor_dir = ((0, -1), (0, 1), (-1, 0), (1, 0))

        @functools.lru_cache(maxsize=None)
        def _check_edge(shape: TileShape, side: TileSide) -> Tuple[bool, bool]:
            """Gets properties of a given edge for a given shape.

            Returns:
                (exists: bool, flush: bool)
            """
            ind = SIDE_CLOCKWISE_INDEX[side]
            vert_a = SHAPE_VERTEXES[shape][ind]
            vert_b = SHAPE_VERTEXES[shape][(ind + 1) & 0x3]

            if abs(vert_a[0] - vert_b[0]) + abs(vert_a[1] - vert_b[1]) <= 1:
                return False, False
            return True, vert_a[0] == vert_b[0] or vert_a[1] == vert_b[1]

        for (layer, x, y), tile in self.tiles.items():
            for side in TileSide:
                edge_exists, edge_flush = _check_edge(tile.shape, side)

                if not edge_exists:
                    tile.edge_data[side] = TileEdgeData()
                    continue

                edat = tile.edge_data[side]
                if not edge_flush:
                    edat.solid = edat.visible = True
                    continue

                # Flush edge, check if neighbor is flush too
                ndir = neighbor_dir[side]
                ntile = self.tiles.get((layer, x + ndir[0], y + ndir[1]))

                if ntile is None:
                    edat.solid = edat.visible = True
                    continue

                nedge_exists, nedge_flush = _check_edge(ntile.shape, side ^ 1)
                if not nedge_exists or not nedge_flush:
                    edat.solid = edat.visible = True
                    continue

                edat.solid = False
                if (tile.sprite_set, tile.sprite_tile, tile.sprite_palette) == (
                    ntile.sprite_set,
                    ntile.sprite_tile,
                    ntile.sprite_palette,
                ):
                    edat.visible = False
                elif visible_callback is None:
                    edat.visible = side in (TileSide.BOTTOM, TileSide.RIGHT)
                else:
                    edat.visible = visible_callback(x, y, side, tile, ntile)

    def calculate_edge_caps(self) -> None:
        """Calculates edge/filth cap flags and angles. This should be called
        after all edge visibilty has been determined (see
        :meth:`calculate_edge_visibility`).

        To compute edge caps we consider only visible edges. Non-visible
        edges will have their cap data appropriately zeroed. For each
        visible edge we consider it in both orientations; going clockwise
        and counter-clockwise around the tile.

        Edges that end between tile widths (i.e. for the slant edge of a
        slant) can never have an edge cap. For these edges the edge/filth cap
        should be set to False and the angles zeroed.

        The first step to computing the cap flag and angle for a given edge
        orientation is to find the "joining" edge. A joining edge must have
        the following properties:

        * Belong to a tile with the same sprite
        * Be the same side of the tile (i.e. ground edges don't connect to walls)
        * Have a starting point equal to our edge's ending point
        * Be in the same orientation as our edge

        If there are multiple joining edges the one that moves the most "away"
        from our tile should be selected (when traversing clockwise the edge
        that goes the most counter-clockwise and vice versa).

        If there is no joining edge the edge cap should be set to True and
        the edge angle should be zeroed. If there is a joining edge the edge
        cap should be set to False and the edge angle should be half the angle
        delta rounded down. Clockwise turns should be positive,
        counter-clockwise turns should be negative.

        Finally if the edge has no filth then the filth cap and angle should be
        zeroed. If there is filth on this edge but not the joining edge, or the
        filth sprites/spike types don't match, the filth cap should be set to
        True and filth angle to 0. Otherwise the filth cap should be False and
        the filth angle should match the edge angle.
        """
        for (layer, x, y), tile in self.tiles.items():
            for side, edge_data in zip(TileSide, tile.edge_data):
                if not edge_data.visible:
                    # Clear all invalid data for invisible tiles.
                    edge_data.caps = (False, False)
                    edge_data.angles = (0, 0)
                    edge_data.filth_caps = (False, False)
                    edge_data.filth_angles = (0, 0)
                    edge_data.filth_sprite_set = TileSpriteSet.NONE_0
                    edge_data.filth_spike = False
                    continue

                # Use temporarily mutable locals to write results
                caps = [False, False]
                angles = [0, 0]
                filth_caps = [False, False]
                filth_angles = [0, 0]

                for dr in range(2):
                    cw_ind = SIDE_CLOCKWISE_INDEX[side]
                    vert_a = SHAPE_VERTEXES[tile.shape][(cw_ind + 1 - dr) & 0x3]
                    vert_b = SHAPE_VERTEXES[tile.shape][(cw_ind + dr) & 0x3]

                    ddirs: Tuple[Tuple[int, int], ...] = ()
                    if vert_b == (0, 0):
                        ddirs = ((-1, 0), (-1, -1), (0, -1))
                    elif vert_b == (2, 0):
                        ddirs = ((0, -1), (1, -1), (1, 0))
                    elif vert_b == (2, 2):
                        ddirs = ((1, 0), (1, 1), (0, 1))
                    elif vert_b == (0, 2):
                        ddirs = ((0, 1), (-1, 1), (-1, 0))
                    else:  # Slants
                        # No caps allowed on slant half edges
                        continue
                    if dr:
                        ddirs = ddirs[::-1]

                    for dx, dy in ddirs:
                        ntile = self.tiles.get((layer, x + dx, y + dy))
                        if (
                            ntile is None
                            or tile.get_sprite_tuple() != ntile.get_sprite_tuple()
                        ):
                            continue

                        nedge_data = ntile.edge_data[side]
                        if not nedge_data.visible:
                            continue

                        nvert_a = SHAPE_VERTEXES[ntile.shape][(cw_ind + 1 - dr) & 0x3]
                        nvert_b = SHAPE_VERTEXES[ntile.shape][(cw_ind + dr) & 0x3]
                        nvert_a = (nvert_a[0] + dx * 2, nvert_a[1] + dy * 2)
                        nvert_b = (nvert_b[0] + dx * 2, nvert_b[1] + dy * 2)
                        if vert_b == nvert_a:
                            break
                    else:
                        # No joiner
                        caps[dr] = True
                        angles[dr] = 0
                        filth_caps[dr] = bool(edge_data.filth_sprite_set)
                        filth_angles[dr] = 0
                        continue

                    delta_angle = (
                        math.atan2(vert_b[1] - vert_a[1], vert_b[0] - vert_a[0])
                        - math.atan2(nvert_b[1] - nvert_a[1], nvert_b[0] - nvert_a[0])
                    ) % (2 * math.pi)
                    if delta_angle > math.pi:
                        delta_angle -= 2 * math.pi

                    angle = -int(round(delta_angle * 180 / math.pi / 2))

                    # Set caps and angles based on joiner
                    caps[dr] = False
                    angles[dr] = angle
                    filth_angles[dr] = 0
                    filth_caps[dr] = False
                    if edge_data.filth_sprite_set:
                        # pylint: disable=undefined-loop-variable
                        if (edge_data.filth_spike, edge_data.filth_sprite_set) == (
                            nedge_data.filth_spike,
                            nedge_data.filth_sprite_set,
                        ):
                            filth_angles[dr] = angle
                        else:
                            filth_caps[dr] = True

                edge_data.caps = tuple(caps)  # type: ignore
                edge_data.angles = tuple(angles)  # type: ignore
                edge_data.filth_caps = tuple(filth_caps)  # type: ignore
                edge_data.filth_angles = tuple(filth_angles)  # type: ignore



"""
Module defining the Dustforce variable representation
"""
import abc
import collections.abc
from enum import IntEnum
from typing import Any, Dict, List, Optional, Tuple, Type


class VariableType(IntEnum):
    """Enumeration of Var type IDs."""

    #: Special code used in the internal format. Null variables cannot be created.
    NULL = 0
    BOOL = 1
    INT = 2
    UINT = 3
    FLOAT = 4
    #: Dustforce strings are actually byte arrays
    STRING = 5
    #: Vec2 is a (float, float) tuple
    VEC2 = 10
    #: Generic mapping/object type
    STRUCT = 14
    ARRAY = 15


class Variable(metaclass=abc.ABCMeta):
    """Variable base class. Variables are a mechanism Dustforce uses to make structure
    metadata easily available to the game script. Variables will raise an `AssertionError`
    if they are created with an invalid :attr:`value` attribute.

    Variables support the equality and hashing interface.

    Attributes:
        value: The internal value of this variable. Its type will depend on the
            actual variable type.
    """

    _TYPES: Dict[VariableType, Type["Variable"]] = {}
    _vtype = VariableType.NULL

    def __init__(self, value: Any) -> None:
        self.value = value
        self._assert_types()

    @classmethod
    def __init_subclass__(cls):
        """record type id to type class mapping"""
        if cls._vtype is not VariableType.NULL:
            Variable._TYPES[cls._vtype] = cls

    def __eq__(self, oth) -> bool:
        # pylint: disable=no-member
        if not isinstance(oth, Variable):
            return False
        return self._vtype == oth._vtype and self.value == oth.value

    def __hash__(self):
        # pylint: disable=no-member
        return hash((self._vtype, self.value))

    def assert_types(self) -> None:
        """Checks if the type of :attr:`value` matches our concrete variable type.

        Raises:
            ValueError: :attr:`value`'s type is invalid
        """
        try:
            self._assert_types()
        except AssertionError as e:
            raise ValueError("value does not match variable type") from e

    @abc.abstractmethod
    def _assert_types(self) -> None:
        """Assert that value types are correct, recursively"""

    def __repr__(self) -> str:
        # pylint: disable=no-member
        return "Variable(%s, %s)" % (repr(self._vtype), repr(self.value))


class VariableBool(Variable):
    """Represents a boolean variable of type :class:`bool`."""

    _vtype = VariableType.BOOL

    def __init__(self, value: bool = False) -> None:
        super().__init__(value)

    def _assert_types(self) -> None:
        """Ensure value is boolean"""
        assert isinstance(self.value, bool)


class VariableInt(Variable):
    """Represents a 32-bit signed int variable of type :class:`int`."""

    _vtype = VariableType.INT

    def __init__(self, value: int = 0) -> None:
        super().__init__(value)

    def _assert_types(self) -> None:
        """Ensure value is int in range."""
        assert isinstance(self.value, int) and -(2 ** 31) <= self.value < 2 ** 31


class VariableUInt(Variable):
    """Represents a 32-bit unsigned int variable of type :class:`int`."""

    _vtype = VariableType.UINT

    def __init__(self, value: int = 0) -> None:
        super().__init__(value)

    def _assert_types(self) -> None:
        """Ensure value is int in range."""
        assert isinstance(self.value, int) and 0 <= self.value < 2 ** 32


class VariableFloat(Variable):
    """Represents a floating point variable of type :class:`float`."""

    _vtype = VariableType.FLOAT

    def __init__(self, value: float = 0.0) -> None:
        super().__init__(value)

    def _assert_types(self) -> None:
        """Ensure value is float."""
        assert isinstance(self.value, float)


class VariableString(Variable):
    """Represents a string variable of type :class:`bytes`."""

    _vtype = VariableType.STRING

    def __init__(self, value: bytes = b"") -> None:
        super().__init__(value)

    def _assert_types(self) -> None:
        """Ensure value is bytes."""
        assert isinstance(self.value, bytes)


class VariableVec2(Variable):
    """Represents a 2-dimensional vector of type `(float, float)`."""

    _vtype = VariableType.VEC2

    def __init__(self, value: Tuple[float, float] = (0.0, 0.0)) -> None:
        super().__init__(value)

    def _assert_types(self) -> None:
        """Ensure tuple of two floats."""
        assert (
            isinstance(self.value, tuple)
            and len(self.value) == 2
            and isinstance(self.value[0], float)
            and isinstance(self.value[1], float)
        )


class VariableStruct(Variable):
    """Represents a struct (dictionary) mapping of type `dict[str, Variable]`."""

    _vtype = VariableType.STRUCT

    def __init__(self, value: Optional[Dict[str, Variable]] = None) -> None:
        super().__init__(dict(value or {}))

    def _assert_types(self) -> None:
        """Ensure value is dictionary of strs to Variables."""
        assert isinstance(self.value, dict)
        for key, val in self.value.items():
            assert isinstance(key, str) and isinstance(val, Variable)
            val._assert_types()


class VariableArray(Variable, collections.abc.MutableSequence):
    """Represents an array variable. Arrays are stored a bit differently
    because they must explicitly encode the type of their sub-elements
    (heterogenous arrays are not allowed).

    :class:`VariableArray` implements the :class:`collections.abc.MutableSequence`
    interface, automatically boxing and unboxing accessed elements.

    Arguments:

        element_type (type[Variable]): Variable type of all elements
        values (list[element_type]): Array of variables of type `element_type`

    Attributes:

        value (element_type, list[element_type]): A tuple containing the element
            type and the element list. Prefer using :attr:`element_type` and
            the :class:`MutableSequence` interface provided by VariableArray instead
            of accessing these elements through :attr:`value`.

    """

    _vtype = VariableType.ARRAY

    def __init__(
        self, element_type: Type[Variable], values: Optional[List[Variable]] = None
    ) -> None:
        super().__init__((element_type, list(values or [])))

    @property
    def element_type(self) -> Type[Variable]:
        """Element type of this array"""
        return self.value[0]

    def _box(self, val):
        """Helper method to box values in a Variable object. Basically just
        VariableArray is weird.
        """
        etype = self.value[0]
        if etype is VariableArray:
            return etype(*val)
        return etype(val)

    # Implement the MutableMapping abstract methods by forwarding to the
    # backing list. We unbox variables from their Variable packaging. This
    # is a little weird with arrays where the unboxed type is a (type, list)
    # tuple but works in principle.
    def __getitem__(self, key):
        return self.value[1][key].value

    def __setitem__(self, key, val):
        self.value[1][key] = self._box(val)

    def __delitem__(self, key):
        del self.value[1][key]

    def __len__(self):
        return len(self.value[1])

    # Method labeled as private since none of the other MutableSequence methods
    # are directly included in the docs.
    def insert(self, index, value):
        """Impelements the :class:`MutableSequence.insert` interface by inserting
        the boxed value into the backing variable list.

        :meta private:
        """
        self.value[1].insert(index, self._box(value))

    def _assert_types(self) -> None:
        """Ensure value is tuple of type and list of variables."""
        assert (
            isinstance(self.value, tuple)
            and len(self.value) == 2
            and isinstance(self.value[0], type)
            and issubclass(self.value[0], Variable)
            and isinstance(self.value[1], list)
        )
        for elem in self.value[1]:
            assert isinstance(elem, self.value[0])
            elem._assert_types()



#!/usr/bin/env python3
"""
Sample script to extract and set level thumbnails.
"""
import argparse
import io
import os
import sys

from dustmaker import DFReader, DFWriter
from dustmaker.cmd.common import (
    run_utility,
    CliUtility,
)
from dustmaker.variable import VariableBool


class Thumbnail(CliUtility):
    """CLI utility for adjusting level thumbnails"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        """Read CLI arguments"""
        parser.description = "extract or update a level thumbnail"
        parser.add_argument("level")
        parser.add_argument("image")
        parser.add_argument(
            "--force",
            action="store_const",
            const=True,
            default=False,
            required=False,
            help="allow overwrite of existing image",
        )
        parser.add_argument(
            "--update",
            action="store_const",
            const=True,
            default=False,
            required=False,
            help="read in the image and update the level thumbnail",
        )
        parser.add_argument(
            "--auto-convert",
            action="store_const",
            const=True,
            default=False,
            required=False,
            help="automatically convert to PNG format (implies --update)",
        )
        parser.add_argument(
            "--auto-scale",
            action="store_const",
            const=True,
            default=False,
            required=False,
            help="automaticaly scale image to expected 382 x 182 size (implies --auto-convert)",
        )

    def main(self, args) -> int:
        """thumbnail CLI entrypoint"""
        if args.auto_scale:
            args.auto_convert = True
        if args.auto_convert:
            args.update = True

        with DFReader(open(args.level, "rb")) as reader:
            level, region_offsets = reader.read_level_ex()
            region_data = b""
            if args.update:
                region_data = reader.read_bytes(region_offsets[-1])

        if not args.update:
            if not args.force and os.path.exists(args.image):
                print("path already exists, use --force to ignore")
                return 1

            with open(args.image, "wb") as fout:
                fout.write(level.sshot)

            return 0

        if args.auto_convert:
            try:
                # pylint: disable=import-outside-toplevel
                from PIL import Image  # type: ignore
            except ImportError:
                print(
                    "failed to import PIL, cannot convert image (try `pip install pillow`)"
                )
                return 1

            with Image.open(args.image) as im:
                if args.auto_scale:
                    im = im.resize((382, 182))
                with io.BytesIO() as io_out:
                    im.save(io_out, format="PNG")
                    level.sshot = io_out.getvalue()
        else:
            with open(args.image, "rb") as fimg:
                level.sshot = fimg.read()

        level.variables["icon_taken"] = VariableBool(True)
        with DFWriter(open(args.level, "wb")) as writer:
            writer.write_level_ex(level, region_offsets, region_data)

        return 0


if __name__ == "__main__":
    sys.exit(run_utility(Thumbnail))



"""
Module defining the core binary reader for Dustforce binary formats.
"""
import io
import itertools
from typing import BinaryIO, Iterable


class BitIO:
    """Wrapper around a binary IO source that allows integers to be read/written
    to. Reads and writes of integers are all serialized to bits in little
    endian order.

    Within the IO source bits are ordered from LSB to MSB. Therefore the
    first bit of a stream is the '1's place of the first byte. The last bit
    of a stream is the '128's place bit of the last byte.

    Arguments:
        noclose (bool): Normally when the BitIO object is closed
            :attr:`data` is also closed. If this is set then :attr:`data`
            will be left open after this BitIO is closed.
    """

    __slots__ = ("data", "_tell", "_bits", "_bits_left", "_noclose")

    def __init__(self, data: BinaryIO, *, noclose: bool = False) -> None:
        #: BinaryIO: A binary data stream. Should support read/write/seek
        #: if those respective operations are done on the BitIO object itself.
        self.data = data
        self._tell = 0
        self._bits = 0
        self._bits_left = 0
        self._noclose = noclose
        try:
            self._tell = data.tell() << 3
        except (AttributeError, IOError):
            pass

    def release(self) -> None:
        """Prevents :meth:`close` from closing :attr:`data` as well."""
        self._noclose = True

    def close(self) -> None:
        """Flush any pending bits and close :attr:`data` unless
        it has been released.
        """
        if not self._noclose:
            self.data.close()

    def aligned(self) -> bool:
        """Returns True if the stream is aligned at a byte boundary."""
        return self._bits_left == 0

    def align(self) -> None:
        """Seeks the stream forward to the nearest byte boundary. This does
        not require :attr:`data` to support seek itself.
        """
        self._tell += self._bits_left
        self._bits = 0
        self._bits_left = 0

    def skip(self, bits: int) -> None:
        """Skips `bits` bits in the bit stream. Requires :attr:`data` to
        support seeks.

        Arguments:
            bits (int): the number of bits to skip
        """
        self.bit_seek(self._tell + bits)

    def bit_tell(self) -> int:
        """Returns the current bit position of the stream"""
        return self._tell

    def bit_seek(self, pos: int) -> None:
        """Seeks to a new bit-position in the stream.

        Arguments:
            pos (int): The position in bits from the start of the stream
        """
        self.data.seek(pos // 8)
        self._tell = pos
        self._bits_left = -pos % 8
        self._bits = 0 if self._bits_left == 0 else -1

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class BitIOReader(BitIO):
    """Bit reader wrapper for a data stream"""

    def __init__(self, data: BinaryIO, *, noclose: bool = False) -> None:
        super().__init__(data, noclose=noclose)

    def read(self, bits: int, signed: bool = False) -> int:
        """Reads in the next `bits` bits into an integer in little endian order.

        Arguments:
            bits (int): The number of bits to read in
            signed (bool): Wether the most significant bit should be interpretted
                as a sign bit.
        """
        bytes_needed = (bits - self._bits_left + 7) >> 3
        if self._bits == -1:
            # We did a seek mid-byte and need to read the lead byte still.
            bytes_needed += 1

        if bytes_needed == 0:
            # No need to actually read
            result = self._bits & (1 << bits) - 1
            self._bits = self._bits >> bits
            self._bits_left -= bits
            self._tell += bits
            return result

        # Read in needed data from underlying stream.
        new_bytes = self.data.read(bytes_needed)

        if self._bits == -1:
            self._bits = new_bytes[0] >> (8 - self._bits_left)
            new_bytes = new_bytes[1:]

        # Calculate the result with maybe some extra data from the last byte.
        result = self._bits + (int.from_bytes(new_bytes, "little") << self._bits_left)

        # Save the extra data from the last byte
        self._bits = result >> bits
        self._bits_left = (self._bits_left - bits) & 7

        # Remove the extra data
        result = result & (1 << bits) - 1
        if signed and (result & (1 << bits - 1)):
            result -= 1 << bits

        self._tell += bits
        return result

    def read_bytes(self, num: int) -> bytes:
        """Reads in the next `num` bytes and returns them as a `bytes` object.

        Arguments:
            num (int): The number of bytes to extract from the bit stream.
        """
        if self._bits_left == 0:
            data = self.data.read(num)
            self._tell += len(data) << 3
            return data
        return bytes(self.read(8) for _ in range(num))


class BitIOWriter(BitIO):
    """Bit writer wrapper for a data stream."""

    def __init__(self, data: BinaryIO, *, noclose: bool = False) -> None:
        if not isinstance(data, io.BufferedIOBase):
            data = io.BufferedWriter(data)  # type: ignore
        super().__init__(data, noclose=noclose)

    def write(self, bits: int, val: int) -> None:
        """Writes `val`, an integer of `bits` bits in size, to the stream.

        Note:
            If the last byte is partially completed it will not be written
            until the stream is closed or flushed.
        """
        if val < 0:
            val += 1 << bits

        if self._bits == -1:
            # Mid seek write will lose data.
            self._bits = 0

        # Fill in the missing bits of the lead byte
        if self._bits_left != 0:
            lead_val = val & (1 << self._bits_left) - 1
            self._bits |= lead_val << (8 - self._bits_left)

        if bits < self._bits_left:
            self._bits_left -= bits
            self._tell += bits
            return

        # Figure out what bytes need to be written to the stream
        full_bytes = (bits - self._bits_left) >> 3
        val_as_bytes = (val >> self._bits_left).to_bytes(full_bytes + 1, "little")
        byte_gen: Iterable[int] = itertools.islice(val_as_bytes, full_bytes)
        if self._bits_left != 0:
            byte_gen = itertools.chain((self._bits,), byte_gen)
        self.data.write(bytes(byte_gen))

        # Fix up our state for leftover bits
        used_bits = (full_bytes << 3) + self._bits_left
        if used_bits == bits:
            self._bits = 0
            self._bits_left = 0
        else:
            self._bits = val_as_bytes[-1]
            self._bits_left = 8 - (bits - used_bits)

        self._tell += bits

    def write_bytes(self, buf: bytes) -> None:
        """Writes the bytes in `buf` to the stream

        Arguments:
            buf (bytes): The data to write to the stream
        """
        if self._bits_left == 0:
            self.data.write(buf)
            self._tell += len(buf) << 3
            return
        for byt in buf:
            self.write(8, byt)

    def close(self) -> None:
        """Flush any pending bits and close the underlying stream
        (unless released).
        """
        try:
            self.flush()
        finally:
            super().close()

    def flush(self) -> None:
        """Flushes any trailing bits.

        Warning:
            If there are trailing bits this will cause the stream to seek
            forward to the next byte boundary. Generally you shouldn't need to
            call this directly and should allow other methods like :meth:`close`,
            :meth:`align`, :meth:`bit_seek` to call flush for you at times that
            always make sense.
        """
        if self._bits_left != 0 and self._bits != -1:
            self.data.write(bytes((self._bits,)))

    def align(self) -> None:
        """Seeks the stream forward to the nearest byte boundary. This does
        not require :attr:`data` to support seek itself. This also triggers
        a flush.
        """
        self.flush()
        super().align()

    # pylint: disable=arguments-differ
    def bit_seek(self, pos: int, *, allow_unaligned: bool = False) -> None:
        """Seeks to the desired position in the stream relative the start.
        This also triggers a flush of any pending data at our current location.

        Arguments:
            pos (int): The bit position to seek to relative the start of the
                stream.
            allow_unaligned (bool): Normally unaligned seeks are not allowed.
                If you set this flag they will be allowed however any write
                performed at the new location will have the effect of zero'ing any
                bits earlier within the byte that we are seeking into.

        Warning:
            Seeking into a non-byte aligned position is not well supported and
            cannot be done generally without performing a read.

        Raises:
            RuntimeError: If seek is not byte aligned and `allow_unaligned` is
                not set.
        """
        if not allow_unaligned and (pos & 7) != 0:
            raise RuntimeError(
                "cannot perform unaligned seek, set allow_unaligned=True if you really want this"
            )
        self.flush()
        super().bit_seek(pos)



""" Dustmaker explicit public exports. This is not the extent of the public
API, merely a convenience for accessing the most commonly used parts.
"""
from .level import Level
from .entity import Entity
from .prop import Prop
from .variable import Variable
from .exceptions import LevelException, LevelParseException
from .tile import Tile

from .dfreader import read_level, DFReader
from .dfwriter import write_level, DFWriter



#!/usr/bin/env python3
"""
Sample script to manipulate level variables.
"""
import argparse
import json
import sys
from typing import Any, Dict, Tuple, Type

from dustmaker import DFReader, DFWriter
from dustmaker.variable import (
    Variable,
    VariableBool,
    VariableStruct,
    VariableInt,
    VariableUInt,
    VariableFloat,
    VariableString,
    VariableVec2,
    VariableArray,
)
from dustmaker.cmd.common import (
    run_utility,
    CliUtility,
)

KNOWN_VAR_FILE_HEADERS = {b"DF_STA", b"DF_CFG", b"DF_FOG"}

_VAR_TYPE_TO_SCHEMA: Dict[Type[Variable], str] = {
    VariableBool: "bool",
    VariableInt: "int",
    VariableUInt: "uint",
    VariableFloat: "float",
    VariableString: "string",
    VariableVec2: "vec2",
}

_SCHEMA_TO_VAR_TYPE: Dict[str, Type[Variable]] = {
    schema: var_type for var_type, schema in _VAR_TYPE_TO_SCHEMA.items()
}


def merge_schema(lhs: Any, rhs: Any) -> Any:
    """Merge two schemas and return the result.

    Arguments:
        lhs (schema): First schema to merge
        rhs (schema): Second schema to merge
    """
    if isinstance(lhs, dict) and isinstance(rhs, dict):
        result = dict(lhs)
        for subkey, subschema in rhs.items():
            r_subschema = result.get(subkey)
            if r_subschema is None:
                result[subkey] = subschema
            else:
                result[subkey] = merge_schema(r_subschema, subschema)
        return result
    if isinstance(lhs, list) and isinstance(rhs, list):
        return [merge_schema(lhs[0], rhs[0])]
    if not isinstance(lhs, str) or not isinstance(rhs, str):
        raise ValueError("unmergable schemas")
    if lhs != rhs:
        raise ValueError(f"differing schemas {lhs} and {rhs}")
    return lhs


def json_to_variables(schema: Any, jvars: Any) -> Variable:
    """Convert JSON variable serialization back into a :class:`Variable`.
    This is the inverse of :meth:`variables_to_json`.

    Arguments:
        schema (schema): The JSON serialized schema as returned from
            :meth:`variables_to_json`
        jvars: The JSON serialized variables as returned from
            :meth:`variables_to_json`

    Returns:
        A :class:`Variable` as the deserialized JSON.
    """
    if isinstance(schema, list):
        element_type: Type[Variable]
        if isinstance(schema[0], list):
            element_type = VariableArray
        elif isinstance(schema[0], dict):
            element_type = VariableStruct
        else:
            element_type = _SCHEMA_TO_VAR_TYPE[schema[0]]

        return VariableArray(
            element_type,
            [json_to_variables(schema[0], elem_jvar) for elem_jvar in jvars],
        )
    if isinstance(schema, dict):
        return VariableStruct(
            {
                key: json_to_variables(schema[key], elem_jvar)
                for key, elem_jvar in jvars.items()
            }
        )
    if schema == "string":
        return VariableString(jvars.encode("latin1"))
    if schema == "vec2":
        return VariableVec2(tuple(jvars))  # type: ignore
    return _SCHEMA_TO_VAR_TYPE[schema](jvars)


def variables_to_json(var: Variable) -> Tuple[Any, Any]:
    """Serialize `var` into a (`schema`, `jvar`) tuple, each of which
    is JSON serializable and human readable.

    Arguments:
        var (Variable): The variable to serialize

    Returns:
        The schema and jvar serialization as a two-tuple.
    """
    if isinstance(var, VariableString):
        return ("string", var.value.decode("latin1"))
    if isinstance(var, VariableVec2):
        return ("vec2", list(var.value))
    if type(var) in _VAR_TYPE_TO_SCHEMA:
        return _VAR_TYPE_TO_SCHEMA[type(var)], var.value

    schema: Any
    if isinstance(var, VariableStruct):
        schema = {}
        jvars = {}
        for subkey, subvar in var.value.items():
            subschema, subjvar = variables_to_json(subvar)
            schema[subkey] = subschema
            jvars[subkey] = subjvar
        return schema, jvars
    if isinstance(var, VariableArray):
        # pylint: disable=redefined-variable-type
        if var.element_type is VariableArray:
            schema = [[]]
        elif var.element_type is VariableStruct:
            schema = [{}]
        else:
            schema = [_VAR_TYPE_TO_SCHEMA[var.element_type]]

        jvar = []
        for elem_var in var.value[1]:
            elem_schema, elem_jvar = variables_to_json(elem_var)
            jvar.append(elem_jvar)
            schema = merge_schema(schema, [elem_schema])
        return schema, jvar

    raise ValueError("unexpected variable type")


class Variables(CliUtility):
    """CLI utility for viewing and modifying level variables"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        """Read CLI arguments"""
        parser.description = (
            "view or modify DF variables as JSON documents. Strings are "
            "encoded/decoded from their internal byte representation using "
            "latin1 encoding."
        )
        parser.add_argument("object")
        parser.add_argument("new_data", nargs="?")
        parser.add_argument(
            "--header",
            default="",
            required=False,
            help="write file from scratch with given header",
        )
        parser.add_argument(
            "--indent",
            default=4,
            required=False,
            type=int,
            help="JSON indent amount in spaces. Use 0 to do no indentation",
        )

    def main(self, args) -> int:
        """thumbnail CLI entrypoint"""
        if args.new_data:
            return self.update_variables(args)
        return self.output_variables(args)

    @staticmethod
    def update_variables(args) -> int:
        """Update the variables in the object requested"""
        with open(args.new_data, "r") as ndata:
            jdata = json.load(ndata)

        if args.header:
            header = args.header.encode()
            if header not in KNOWN_VAR_FILE_HEADERS:
                sys.stderr.write(f"cannot write {header} file")
                return 1
        else:
            with DFReader(open(args.object, "rb")) as reader:
                header = reader.read_bytes(6)
                reader.bit_seek(0)
                if header == b"DF_LVL":
                    level, region_offsets = reader.read_level_ex()
                    region_data = reader.read_bytes(region_offsets[-1])
                elif header not in KNOWN_VAR_FILE_HEADERS:
                    sys.stderr.write(f"unknown file header {header}\n")
                    return 1

        variables = json_to_variables(jdata["schema"], jdata["vars"])

        with DFWriter(open(args.object, "wb")) as writer:
            if header == b"DF_LVL":
                level.variables = variables.value
                writer.write_level_ex(level, region_offsets, region_data)
            else:
                writer.write_var_file(header, variables.value)

        return 0

    @staticmethod
    def output_variables(args) -> int:
        """Output the variables of the object requested"""
        with DFReader(open(args.object, "rb")) as reader:
            header = reader.read_bytes(6)
            reader.bit_seek(0)
            if header == b"DF_LVL":
                level = reader.read_level(metadata_only=True)
                variables = level.variables
            elif header in KNOWN_VAR_FILE_HEADERS:
                variables = reader.read_var_file(header)
            else:
                sys.stderr.write(f"unknown file header {header}\n")
                return 1

        schema, jvars = variables_to_json(VariableStruct(variables))

        json.dump(
            {
                "vars": jvars,
                "schema": schema,
            },
            sys.stdout,
            indent=args.indent if args.indent else None,
        )

        return 0


if __name__ == "__main__":
    sys.exit(run_utility(Variables))



""" Shared CLI definitions """
import abc
import argparse
from typing import Type


class CliUtility(metaclass=abc.ABCMeta):
    """Abstract utility base class"""

    @abc.abstractmethod
    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        """Allow the utility to set up its arg parser.

        Arguments:
            parser (argparse.ArgumentParser): the parser to setup arguments for.
        """

    @abc.abstractmethod
    def main(self, args) -> int:
        """Run utility with parsed args.

        Returns:
            The utility exit code
        """


def run_utility(utility: Type[CliUtility]) -> int:
    """Run a single command line utility.

    Arguments:
        utility (type[CliUtilitiy]): class of utility to run.

    Returns:
        The desired exit code
    """
    util = utility()
    parser = argparse.ArgumentParser()
    util.setup_parser(parser)
    return util.main(parser.parse_args())
